Require Import Rdefinitions.
Require Import ssreflect ssrbool eqtype ssrnat seq fintype bigop ssralg.
Require Import poly matrix mxpoly mxalgebra ssrnum complex zmodp polydiv.
Require Import Rstruct C_complements set_theory big_minmax.
Require Import mxorder struct_top.
Require Import ssrcomplements mxstructure similar closed_poly.
Require Import frobenius_form jordan.

(******************************************************************************)
(*                                                                            *)
(*    This file contains a définitions and some theory about the spectral     *)
(*    radius of matrices.                                                     *)
(*                                                                            *)
(*                                                                            *)
(*           eigen_seq M == the sequence of all eigenvalue of M.              *)
(*                \rho M == the spectral radius of the marix M.               *)
(******************************************************************************)
 


Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

Import GRing.Theory.
Import Num.Theory.
Import Num.Def.

Section C_archi.

Local Notation C := (complex Rdefinitions.R).
Open Scope ring_scope.

Lemma archiC_Mixin : 
 Num.archimedean_axiom [numDomainType of C].
Proof.
move=> x; exists (Num.bound (Re `|x|)).
by rewrite -natC -[`|_|]ReK // ltcR archi_boundP // Re_ge0.
Qed.

Definition C_archiDomain := ArchiDomainType C archiC_Mixin. 
Canonical C_archiDomain.

End C_archi. 

Section R_Top.

Let Rn := [numDomainType of R].

Canonical Structure ArchiD_R := Eval hnf in 
  [archiDomainType of R for (archi_archi [archiFieldType of R])]. 
Canonical Structure EspMet_R := 
  Eval hnf in [espMetType R of R for (numEspMetType Rn)].  
Canonical Structure EspTop_R := 
  Eval hnf in [topologyType of R for (numTopologyType Rn)].

End R_Top.

Section definition.

Variable R : rcfType.
Local Notation C := (complex R).

Open Scope ring_scope.

Local Notation "A ^%:C" := (map_mx (real_complex_def (Phant R)) A) (at level 2).

Definition eigen_seq n (M : 'M[C]_n) := root_seq (char_poly M).

Definition spectral_radius n (M : 'M[C]_n) :=
  (\big[(@maxr _)/0]_(lam <- eigen_seq M) Re `|lam|).

Notation "\rho A" := (spectral_radius A) (at level 0).

Lemma eigenE n (M : 'M[C]_n) lam : 
 (lam \in eigenvalue M) = (lam \in eigen_seq M).
Proof.
rewrite -topredE /= eigenvalue_root_char.
by rewrite -root_root_seq // -size_poly_eq0 size_char_poly.
Qed.

Lemma rho_ge_eigen n (M : 'M[C]_n) lam : lam \in eigenvalue M -> 
  `|lam| <= (\rho M)%:C.
Proof.
rewrite eigenE => Hlam.
rewrite /spectral_radius -[`|_|]ReK // lecR.
have Hge0 la : 0 <= Re `|la| by rewrite Re_ge0. 
exact: (bigmax_ger (Hge0 _) Hlam).
Qed.

Lemma eigen_seq_nnil n (M : 'M[C]_n.+1) : eigen_seq M != [::].
Proof. by rewrite -size_eq0 size_root_seq size_char_poly. Qed.

Lemma rho0 n : \rho (0 : 'M_n) = 0.
Proof.
rewrite /spectral_radius; apply/eqP.
  apply/eqP/bigmax_eq0=> i.
  rewrite -eigenE=> /eigenvalueP [x] /eqP.
  rewrite mulmx0 eq_sym scaler_eq0=> /orP []=>[/eqP ->|->] //.
  by rewrite normr0.
Qed.

Lemma rho_ge0 n (M : 'M_n) : 0 <= \rho (M).
Proof.
rewrite /spectral_radius /eigen_seq.
case: n M=>[M|n M].
  have/eqP ->: root_seq (char_poly M) == [::].
    by rewrite -root_seq_nil size_char_poly.
  by rewrite big_nil.
have Hnv: root_seq (char_poly M) != [::].
    by rewrite -root_seq_nil size_char_poly.
apply: bigmax_pos=> i.
by rewrite Re_ge0.
Qed.

Lemma char_poly_trmx (T : comRingType) n (A : 'M[T]_n) :
  char_poly A = char_poly A^T.
Proof.
by rewrite /char_poly /char_poly_mx -map_trmx -{2}tr_scalar_mx -linearB det_tr.
Qed.

Lemma eigen_trmx (F : fieldType) n (A : 'M[F]_n) lam :
  lam \in eigenvalue A = (lam \in eigenvalue (A^T)).
Proof. by rewrite -!topredE /= !eigenvalue_root_char char_poly_trmx. Qed.

Lemma eigenvalue2P (F : fieldType) n (A : 'M[F]_n) lam :
  reflect (exists2 v : 'cV_n, A *m v = lam *: v & v != 0) (lam \in eigenvalue A ).
Proof.
rewrite eigen_trmx -topredE /=.
apply: (iffP (@eigenvalueP F n (A^T) lam))=> [][v Hv1 Hv2]; exists (v ^T).
  +by rewrite -linearZ -Hv1 /= trmx_mul trmxK.
  +by rewrite raddf_eq0 //; apply: trmx_inj.
-by rewrite -linearZ -Hv1 /= trmx_mul.
-by rewrite raddf_eq0 //; apply: trmx_inj.
Qed.

Lemma ex_eigen_rho  n (A : 'M[C]_n.+1) : 
  {lam | eigenvalue A lam & `|lam| = \rho(A)%:C}.
Proof.
have Hgt0 lam : (0 : R) <= Re `|lam| by rewrite Re_ge0.
have [k /andP []] := bigmax_eq Hgt0 (eigen_seq_nnil A).
rewrite -eigenE=> ? /eqP Hk.
by exists k=> //; rewrite /spectral_radius -Hk ReK. 
Qed.

Lemma rho_scalemx n a (A :'M_n) :  \rho (a *: A) = Re `|a| * \rho A.
Proof.
case: n A => [A |n A].
  have ->: A = 0 by apply/matrixP=> [][].
  by rewrite scaler0 rho0 mulr0.
case: (altP (a=P0))=> [->|Hanz].
  by rewrite scale0r normr0 rho0 mul0r.
have Hgt0 lam : (0 : R) <= Re `|lam| by rewrite Re_ge0.  
rewrite /spectral_radius.
have [k /andP[]]:= bigmax_eq Hgt0 (eigen_seq_nnil A).
rewrite -eigenE=> /eigenvalueP [v Hv Hvneq0] /eqP Hk.
have -> //:  \maxr_(lam <- eigen_seq (a *: A)) Re `|lam| =
        Re `|a| * \maxr_(lam <- eigen_seq A) Re `|lam|.
  apply/eqP. rewrite eqr_le ; apply/andP; split; last first.
    have: a * k \in (eigenvalue (a *: A)).
    by apply/eigenvalueP; exists v=> //; rewrite -scalemxAr Hv scalerA.
    rewrite eigenE=> /(bigmax_ger Hgt0).
     rewrite -Hk.
     by rewrite normrM real_ReM.  
  have [k' /andP[]]:= bigmax_eq Hgt0 (eigen_seq_nnil (a *: A)).
  rewrite -eigenE=> /eigenvalueP [v' Hv' Hv'neq0] /eqP <-.
  have: a^-1 * k' \in eigenvalue A.
    apply/eigenvalueP; exists v'=> //;  apply: (scalerI Hanz). 
    by rewrite scalemxAr Hv' scalerA mulrA mulrV ?mul1r.
  rewrite eigenE => /(bigmax_ger Hgt0). 
  rewrite normrM real_ReM // normrV // real_ReV // ler_pdivr_mull //.
  by rewrite ltr_neqAle Hgt0 eq_sym real_Re_eq0 // normr_eq0 Hanz.
Qed.

End definition.

Notation "\rho A" := (spectral_radius A) (at level 0).
 
Section perron_thm.

Local Notation C := (complex R).
Local Notation "A ^%:C" := (map_mx (real_complex_def (Phant R)) A) (at level 2).
Local Notation "A ^%:R" := (map_mx (@Re _) A) (at level 2).

Open Scope ring_scope.

Lemma root_invf_eigen (F: fieldType) m (A : 'M[F]_m) (p : {poly F}) x : 
p \in (invariant_factors A) -> root p x -> eigenvalue A x.
Proof.
move=> Hp.
rewrite eigenvalue_root_char -!dvdp_XsubCl=> Hr.
by apply/(dvdp_trans Hr)/dvdp_invf_char_poly.
Qed.

Lemma mx_cvg0P n (A :'M[R]_n.+1) : 
  reflect ((fun k => A^+k) >->> 0) (\rho (A^%:C) < 1).
Proof. 
apply/(iffP idP)=>[Hrho|].
  have [eq [P [HP HJ]]] := Jordan (A ^%:C).
  have HA:  A ^%:C = P^-1 *m conform_mx P (Jordan_form A ^%:C) *m P.  
    by rewrite -mulmxA -HJ mulmxA mulVmx // mul1mx. 
  have cvg0: (fun k => (conform_mx P (Jordan_form A ^%:C))^+k) >->> 0.
    apply/(cvg_ext (conform_mxX P _)).
    move: eq; set m := (X in _ = X)=> /(@sym_eq _ _ _) eq.  
    have -> : 0 = (conform_mx P (0: 'M_m)).
      by rewrite eq conform_mx_id .    
    apply: conform_mx_cvg=> //.
    apply/cvg_shiftS/(cvg_ext (exp_diag_block_S _ _)).
    set s := [seq _ | _ <- _].
    have <- : diag_block_mx s (fun _ _=> 0) = 0.
      by move=> t; apply/diag_block_mx0.
    apply: diag_block_mx_cvg=> i Hi.
    apply/cvg_shiftS/cvg_jordan_lt1.
    set lam := (X in `|X|).
    have/rho_ge_eigen lerho : lam \in (eigenvalue (A^%:C)).
      rewrite size_map in Hi.
      have/flattenP [sp /(nthP [::]) [j]] := mem_nth (0,0%N) Hi.
      rewrite size_map=> Hj.
      rewrite (nth_map 0) // => <- /(nthP (0,0%N)) [k].
      rewrite size_map=> Hk.
      rewrite (nth_map 0) // /lam => <- /=.
      set p := _`_j.
      have Hpif := mem_nth 0 Hj.     
      apply: (root_invf_eigen Hpif).
      have := mem_nth 0 Hk.
      by rewrite mem_undup root_root_seq // (invariant_factor_neq0 Hpif).
    apply: (ler_lt_trans lerho).
    by rewrite ltcR. 
  apply/cvg_realc_mx.  
  have ext k : (A ^+ k)^%:C = (A^%:C)^+k.
    elim: k=>[|k IHk]; first by rewrite !expr0 map_mx1.
    by rewrite !exprS map_mxM IHk.
  apply/(cvg_ext ext); rewrite HA -(mul0mx _ P) -(mulmx0 _ P^-1)=> {ext}.
  have ext M k: (P^-1 *m M *m P)^+k = P^-1 *m M^+k *m P. 
    elim: k=>[|k IHk]; first by rewrite !expr0 mulmx1 mulVmx.
    rewrite !exprS IHk /GRing.mul /= -!mulmxA [P *m (_ *m _)]mulmxA mulmxV //.
    by rewrite mul1mx.
  apply/(cvg_ext (ext _))/(continue_cvg (continu_mull (A := P))).  
  exact:(continue_cvg (continu_mulr (A := P^-1))).  
move=> H.
have [lam Hl Hrho]:= ex_eigen_rho A^%:C. 
have/eigenvalueP [v Hv Hnvz]:= Hl.
have Hexp (k : nat): lam^+k *: v = v *m (A^%:C)^+k.
  elim: k; first by rewrite !expr0 mulmx1 scale1r.
  move=> k IHk.
  by rewrite !exprSr mulmxA -scalerA -Hv scalemxAl IHk.
have/(cvg_scale0r Hnvz)/cvg_lt1 Hlam0: (fun k => lam^+k *: v) >->> 0.
  apply:(cvg_ext Hexp).
rewrite -(mulmx0 _ v) .  
apply: (continue_cvg (continu_mulr (A:= v))).  
pose t := GRing.exp A ^%:C 3.
have ext k : (GRing.exp A ^%:C k) = (A ^+ k)^%:C.
  by rewrite rmorphX.  
by apply/(cvg_ext ext)/cvg_realc_mx.
by rewrite -ltcR -Hrho.
Qed.

Lemma PF_gt0 n (A : 'M[R]_n) (x : 'cV_n) (lam : C) : 
  let x' := (Mabs x)^%:R in 
  0 <m: A ->
  x != 0 -> A^%:C *m x = lam *: x -> `|lam| = (\rho (A^%:C))%:C -> 
  [/\ (A *m x' = \rho (A^%:C) *: x'), 0 <m: x' & 0 < \rho (A^%:C)].
Proof.
case: n A x => [A x|n A x x' HAgt0 Hxneq0 Hegv Hlam].
  have ->: x = 0 by apply/matrixP=> [][].
  by rewrite eqxx.
have Hx'ge0 : 0 <=m: x' by move=> i j; rewrite !mxE Re_ge0.
have Heqxx' : Mabs x = x' ^%:C by apply/matrixP=> i j; rewrite !mxE ReK.
have x'neq0 : x' != 0 by rewrite /is_true -Hxneq0 -Mabs_eq0 Heqxx' map_mx_eq0.
have/subr_Mge0 := MabsM A^%:C  x. 
rewrite Heqxx' Mabs_real_complex -!map_mxM Hegv Mabs_scalemx Hlam.
rewrite Heqxx' -map_mxZ -map_mx_sub. 
rewrite ![Mabs A]eq_Mabs_id //; last exact: MltW.
set y := (X in _ <=m: X^%:C)=> /mxreal_ge0 Hge0.
set z := _ *m _.
have Hzgt0 : 0 <m: z.
  by apply (mulmx_gt0 HAgt0 Hx'ge0).
case: (altP (y =P 0))=> [/eqP Hy|Hy].
 have:= rho_ge0 (A^%:C); rewrite ler_eqVlt /y.
  case/orP=> [/eqP Hrho0|rhogt0].
    rewrite /y -Hrho0 scale0r subr0 in Hy. 
    rewrite /z (eqP Hy) in Hzgt0.
    by have:= Hzgt0 0 0; rewrite ltrr.
  rewrite /y subr_eq0 in Hy. 
  split=> // [|i j]; first by apply/eqP.
  rewrite !mxE ltr_neqAle Re_ge0 // normr_ge0 andbT.
  apply/eqP=> Habs.
  have/eqP/matrixP/(_ i j) := Hy.
  rewrite ![in X in _ = X]mxE -Habs mulr0=> H0.
  by have/(_ i j):= Hzgt0; rewrite mxE /z H0 ltrr.
have := mulmx_gt0 HAgt0 Hge0 Hy.
rewrite /y mulmxBr -scalemxAr=> Hpos. 
have/subr_Mgt := Mle_eps Hpos Hzgt0.
set eps := min_fT _.
have Hepsgt0 : 0 < eps.
  rewrite /eps;   set F := (fun i => _).
  have [k ->] := ex_min_fT F ord0.
  have/(_ k 0):= Hzgt0; rewrite mxE=> Hzk0.
  have/(_ k 0):= Hpos; rewrite mxE=> HAz0.
  by rewrite !mulr_gt0 // invr_gt0 // ltr0n.
have Hrgt0 : 0 < (\rho (A^%:C) + eps)^-1.
  by rewrite invr_gt0 -[0]addr0 ler_lt_add // rho_ge0.
rewrite -scalerDl=> /(Mlt_scaler Hrgt0).
have:= Hrgt0; rewrite lt0r -unitfE unitrV=> /andP [Hru _].
rewrite scalerA mulVr // scale1r scalemxAl.
set B := _ *: _=> HB.
have := Hrgt0; rewrite invr_gt0=> Hrgt02.
have/mx_cvg0P rBlt1 : \rho (B^%:C) < 1.
  rewrite map_mxZ rho_scalemx ?ltrW // gtr0_norm ?ltcR //. 
  rewrite -(ltr_pmul2l Hrgt02) ?mulrA mulrV // ?mulr0 ?ltr01 //.
  by rewrite mulr1 mul1r ltr_addl.
have HBltz k : z <m: B^+k.+1 *m z.
  have Htr m : B^+m *m z <m: B^+m.+1 *m z.
    elim: m=>[|m IHm]; first by rewrite mul1mx.
    rewrite 2!exprS -!mulmxA; apply: (Mlt_pmul2l _ IHm).
    move=> i j; rewrite !mxE mulr_gt0 ?invr_gt0 //.
    by have/(_ i j):= HAgt0; rewrite !mxE.
  by elim: k=> // k IHk; apply/(Mlt_trans IHk).
have/continue_cvg Hcvgm := (continu_mull (m:= n.+1) (A:= z)).
have/Hcvgm : (fun k : nat => B ^+ k.+1) >->> 0.
  have Hg: forall m n, (m < n)%N -> (m.+1 < n.+1)%N by [].
  pose g := mkss Hg.
  exact: (cvg_ss_cvg g rBlt1).
rewrite mul0mx=> Hcv0.
have Habs := ltmx_cvg_lemx HBltz Hcv0.
have: 0 < z 0 0 <= 0.
  have/(_ 0 0):= Hzgt0; rewrite mxE=> ->.
  by have/(_ 0 0):= Habs; rewrite [X in _ <= X]mxE=> ->.
by rewrite lter_anti.
Qed.

Lemma Perron_gt0 n (A : 'M[R]_n.+1) : 0 <m: A ->
  exists x : 'cV_n.+1, 0 <m: x /\ A *m x = (\rho (A^%:C)) *: x.
Proof.
move=> HAgt0.
have [lam /eigenvalue2P [x Hx Hx0] Hrho] := ex_eigen_rho (A^%:C).
have [HA Hgt0 _] := PF_gt0 HAgt0 Hx0 Hx Hrho.
by exists (Mabs x)^%:R.
Qed.  

End perron_thm.

