Require Import Classical_Prop IndefiniteDescription Classical_Pred_Type.
Require Import ssreflect ssrfun ssrbool eqtype ssrnat seq choice.
Require Import fintype bigop binomial ssralg matrix ssrnum complex.
Require Import mxstructure ssrcomplements jordan.
Require Import C_complements big_minmax set_theory mxorder.

(******************************************************************************)
(*                                                                            *)
(*    We define in this file an abstract structure of topological space       *)
(*    and of metric space. After we instanciate a metric and a topology :     *)
(*      -on each structure define in the file ssrnum.v of SSReflect library   *)
(*      -on matrices.                                                         *)
(*                                                                            *)
(* This file contains also some general theory about the above structures.    *)
(* In particular he contains the Bolzano-Weirstrass theorem.                  *)
(*                                                                            *)
(*  *** ArchiDomain structure                                                 *)
(*                archiDomainType == it is not exactly an archimedean field.  *)
(*                                   It is more precisely an numFieldType     *)
(*                                     that verify the archimedean axiom.     *)
(*            ArchiDomainType T r == packs the archimedean axiom r into       *)
(*                                   an archiDomainType. The carrier T must   *)
(*                                   have a numFieldtype structure.           *)
(*  [archiDomainType of T for S ] == T-clone of the archiDomainType           *)
(*                                   structure S.                             *)
(*         [archiDomainType of T] == clone of a canonical archiDomainType     *)
(*                                   structure on T.                          *)
(*                                                                            *)
(*  *** Topological space                                                     *)
(*                   topologyType == interface for an topological space.      *)
(*               TopologyType T m == packs the topology mixin into an         *)
(*                                   topologyType.                            *)
(*     [topologyType of T for S ] == T-clone of the topologyType structure S. *)
(*            [topologyType of T] == clone of a canonical topologyType        *)
(*                                   structure on T.                          *)
(*                                                                            *)
(*                                                                            *)
(*                         open A == A is an open set.                        *)
(*                             A° == the interior of the set A.               *)
(*                        close A == A is a close set.                        *)
(*                             /A == the closure of A.                        *)
(*                     boundary A == the boundary of the set A.               *)
(*              neighbourhood V x == V is a neighbourhood of the point x.     *)
(*                 neigh_open V x == V is an open neighbourhood of x.         *)
(*                    separated T == the topologcal space T is separated.     *)
(*                  open_family f == all set of the family sets f are open.   *)
(*                quasi_compact K == K is quasi compact (compacity in a not   *)
(*                                   separated space).                        *)
(*                      compact K == K is a compact set.                      *)
(*                                                                            *)
(*                      Un >->> l == the sequence Un (of type nat -> T,       *)
(*                                   where T is a topological space)          *)
(*                                   converge to l.                           *)
(*             cluster_point Un l == l is a cluster point of the sequence Un. *)
(*                                                                            *)
(*                continue_pt f x == the function f is continue at point x.   *)
(*                     continue f == the function f is continue.              *)
(*                                                                            *)
(*                                                                            *)
(*  *** Metric space                                                          *)
(*                   espMetType R == interface for an metric space where      *)
(*                                   the distance has return type R.          *)
(*               EspMetType R T m == packs the metric mixin into an           *)
(*                                   espMetType.                              *)
(*     [espMetType R of T for S ] == T-clone of the espMetType structure S.   *)
(*            [espMetType R of T] == clone of a canonical espMetType          *)
(*                                   structure on T.                          *)
(*                                                                            *)
(*                                                                            *)
(*                       {posD T} == the type of positive element of          *)
(*                                   numDomainType T.                         *)
(*                           PosD == the constructor of type {posD T}.        *)
(*                                   he take in argument a proof that         *)
(*                                   a number is positive.                    *)
(*                           sort == it is the projection of {posD T} to T.   *)
(*                       lt_pos r == a proof that sort r is positive.         *)
(*                        pceil r == the ceiling of (r : {posD T}).           *)
(*                       pfloor r == the floor of (r : {posD T}).             *)
(*                                                                            *)
(*                                                                            *)
(*                        'B(x,r) == the open ball of center x : E and of     *)
(*                                   radius r : {posD R}, where E has type    *)
(*                                   espMetType R                             *)
(*                      bounded A == the set A is bounded.                    *)
(*                                                                            *)
(*                                                                            *)
(*  *** Matrix norm                                                           *)
(*                       can_norm == a structure for generic canonical norm   *)
(*                                   over matrices.                           *)
(*                        norm8 M == the infinity norm of M.                  *)
(*                    ||M : norm| == the canonical norm "norm" of the         *)
(*                                   matrix M. Here norm is an instance of    *)
(*                                   can_norm structure.                      *)
(******************************************************************************)
 
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Reserved Notation "{ 'posD' T }" (at level 0, format "{ 'posD'  T }").

Import GRing.Theory.
Import Num.Theory.
Import Num.Def.


Module ArchimedeanDomain.

Local Open Scope ring_scope.

Section ClassDef.
Local Notation num_for T b := (@Num.NumDomain.Pack T b T).
Record class_of R :=
  Class { base : Num.NumField.class_of R; 
             _ : Num.archimedean_axiom (num_for R base) }.
Local Coercion base : class_of >-> Num.NumField.class_of.

Structure type := Pack {sort; _ : class_of sort; _ : Type}.
Local Coercion sort : type >-> Sortclass.
Variables (T : Type) (cT : type).
Definition class := let: Pack _ c _ as cT' := cT return class_of cT' in c.
Let xT := let: Pack T _ _ := cT in T.
Notation xclass := (class : class_of xT).
Definition clone c of phant_id class c := @Pack T c T.
Definition pack b0 (m0 : Num.archimedean_axiom (num_for T b0)) :=
  fun bT b & phant_id (Num.NumField.class bT) b =>
  fun    m & phant_id m0 m => Pack (@Class T b m) T.

Definition eqType := @Equality.Pack cT xclass xT.
Definition choiceType := @Choice.Pack cT xclass xT.
Definition zmodType := @GRing.Zmodule.Pack cT xclass xT.
Definition ringType := @GRing.Ring.Pack cT xclass xT.
Definition comRingType := @GRing.ComRing.Pack cT xclass xT.
Definition unitRingType := @GRing.UnitRing.Pack cT xclass xT.
Definition comUnitRingType := @GRing.ComUnitRing.Pack cT xclass xT.
Definition idomainType := @GRing.IntegralDomain.Pack cT xclass xT.
Definition numDomainType := @Num.NumDomain.Pack cT xclass xT.
Definition fieldType := @GRing.Field.Pack cT xclass xT.
Definition numFieldType := @Num.NumField.Pack cT xclass xT.

End ClassDef.

Module Exports.
Coercion base : class_of >-> Num.NumField.class_of.
Coercion sort : type >-> Sortclass.
Bind Scope ring_scope with sort.
Coercion eqType : type >-> Equality.type.
Canonical eqType.
Coercion choiceType : type >-> Choice.type.
Canonical choiceType.
Coercion zmodType : type >-> GRing.Zmodule.type.
Canonical zmodType.
Coercion ringType : type >-> GRing.Ring.type.
Canonical ringType.
Coercion comRingType : type >-> GRing.ComRing.type.
Canonical comRingType.
Coercion unitRingType : type >-> GRing.UnitRing.type.
Canonical unitRingType.
Coercion comUnitRingType : type >-> GRing.ComUnitRing.type.
Canonical comUnitRingType.
Coercion idomainType : type >-> GRing.IntegralDomain.type.
Canonical idomainType.
Coercion numDomainType : type >-> Num.NumDomain.type.
Canonical numDomainType.
Coercion fieldType : type >-> GRing.Field.type.
Canonical fieldType.
Coercion numFieldType : type >-> Num.NumField.type.
Canonical numFieldType.
Notation archiDomainType := type.
Notation ArchiDomainType T m := (@pack T _ m _ _ id _ id).
Notation "[ 'archiDomainType' 'of' T 'for' cT ]" := (@clone T cT _ idfun)
  (at level 0, format "[ 'archiDomainType'  'of'  T  'for'  cT ]") : form_scope.
Notation "[ 'archiDomainType' 'of' T ]" := (@clone T _ _ id)
  (at level 0, format "[ 'archiDomainType'  'of'  T ]") : form_scope.
End Exports.

End ArchimedeanDomain.
Export ArchimedeanDomain.Exports.

Module Import Internals.

Fact archiD_bound_subproof (R : archiDomainType) : Num.archimedean_axiom R.
Proof. by case: R => ? []. Qed.

End Internals.

Module Import ExtraDef.

Definition archiD_bound {R} x := sval (sigW (@archiD_bound_subproof R x)).

End ExtraDef.

Notation boundD := archiD_bound.

Section Theory.

Section ArchimedeanFieldTheory.

Local Open Scope ring_scope.

Variables (F : archiDomainType) (x : F).

Lemma archiD_boundP : 0 <= x -> x < (boundD x)%:R.
Proof. by move/ger0_norm=> {1}<-; rewrite /boundD; case: (sigW _). Qed.

End ArchimedeanFieldTheory.

Lemma archi_axiom (T : archiFieldType) : Num.archimedean_axiom T.
Proof. by case: T=> ? []. Qed.

Canonical Structure archi_archi (T : archiFieldType) := 
  ArchiDomainType T (@archi_axiom T).

End Theory.

Module Topology.

 Structure mixin_of (T_Space : Type) : Type := Mixin
  {
    open : (T_Space -> Prop) -> Prop;
    _ : open {ø : T_Space};
    _ : open {full : T_Space};
    _ : forall (E : (T_Space -> Prop) -> Prop), E <s open -> 
          open (union_s E id);
    _ : forall A B, open A -> open B -> open (inter A B);
    _ : forall A B, A =s B -> (open A <-> open B)
  }.

Section ClassDef.

Notation class_of := mixin_of (only parsing).

Structure type := Pack {sort; _ : class_of sort; _ : Type}.
Local Coercion sort : type >-> Sortclass.

Variables (T : Type) (cT : type).
Definition class := let: Pack _ c _ as cT' := cT return class_of cT' in c.
Definition pack c := @Pack T c T.
Definition clone := fun c  & cT -> T & phant_id (pack c) cT => pack c.

End ClassDef.

Module Exports.

Coercion sort : type >-> Sortclass.

Notation topologyType := type.
Notation TopologyMixin := Mixin.
Notation TopologyType T m := (@pack T m).
Notation " [ 'topologyType' 'of' T ] " := (@clone T _ _ id idfun)
  (at level 0, format "[ 'topologyType' 'of' T ]") : form_scope.
Notation "[ 'topologyType' 'of' T 'for' C ]" := (@clone T C _  id idfun)
  (at level 0, format "[ 'topologyType' 'of' T 'for' C ]") : form_scope.

End Exports.
End Topology.
Export Topology.Exports.

Section TopologyTheory.

Variable T : topologyType.

Canonical Structure Top_eqType := Eval hnf in EqType T (T_eqMixin T).

Definition open := Topology.open (Topology.class T).

Add Morphism open with signature (@eqset T) ==> iff as open_mor.
Proof. by rewrite /open; case: T => ? []. Qed.

Lemma empty_in_open : open {ø : T}.
Proof. by rewrite /open; case: T => ? []. Qed.

Lemma full_in_open : open {full : T}.
Proof. by rewrite /open; case: T => ? []. Qed.

Lemma union_open : forall (E : (T -> Prop) -> Prop), E <s open -> 
      open (union_s E (@id (T ->Prop))).
Proof. by rewrite /open; case: T => ? []. Qed.

Lemma inter_open : forall A B, open A -> open B -> open (inter A B).
Proof. by rewrite /open; case: T => ? []. Qed.

Lemma union_open2 (E: (T -> Prop) -> Prop) F : 
  (forall A, A 'in E -> open (F A)) -> open (union_s E F). 
Proof.
move=> H; rewrite eq_union; apply: union_open=> S.
by case=> A [/ H HA1 HA2]; rewrite -HA2. 
Qed.

Definition interior (A : T -> Prop) := 
  union_s (fun B => open B /\ B <s A) id. 

Notation "A °" := (interior A) (at level 0, no associativity).

Lemma int_go A B : open A -> A <s B -> A <s B°.
Proof. by move=> Ho HAB x Hx; exists A. Qed.

Lemma open_int A : open A°.
Proof. by apply: union_open=> X []. Qed.

Lemma int_in A : A° <s A.
Proof. by move=> x [] E [[H1 H2] /H2 H3]. Qed.

Add Morphism interior with signature (@eqset T) ==> (@eqset T)
  as interior_mor.
Proof.
move=> A B HAB; apply/double_i; apply/int_go; try exact: open_int.
  by rewrite -HAB; apply: int_in.
by rewrite HAB; apply: int_in.
Qed.

Lemma int_open_eq A : open A <-> A <s A°.  
Proof.
split=> H; first exact: int_go. 
have -> : A =s A° by apply: double_i => //; apply: int_in.
exact: open_int.
Qed.

Lemma int_inter A B : (inter A B)° =s inter A° B°.
Proof.
split; case; [|case]=> S [[HS1 HS2] HS3]. 
  by split; exists S; do 2! split=> //; apply/(include_trans HS2)=> y [].
case=> U [[HU1 HU2] HU3].
exists (inter S U); split; split=> //.
  exact: inter_open.
exact: in_inter.
Qed.

Lemma int_union A B : union A° B° <s (union A B)° .
Proof.
move=> x [][] S [[HS1 HS2] HS3]; exists S; do 2! split=>//;
 apply/(include_trans HS2).
  exact: union_left.
exact: union_right.
Qed.

Definition close (A : T -> Prop) : Prop := open (A^c).

Add Morphism close with signature (@eqset T) ==> iff as close_mor. 
Proof. by move=> A B HAB; rewrite /close HAB. Qed.

Lemma empty_close : close {ø : T}.
Proof. by rewrite /close empty_c; apply: full_in_open. Qed.

Lemma full_close : close {full : T}.
Proof. by rewrite /close full_c; apply: empty_in_open. Qed.

Lemma inter_close (E : (T -> Prop) -> Prop) : 
  (forall A, A 'in E -> close A) -> close (inter_s E id).
Proof. by rewrite /close inter_compl_union; apply: union_open2. Qed.  

Lemma union_close A B : close A -> close B -> close (union A B). 
Proof. by move=> HA HB; rewrite /close compl_union; apply: inter_open. Qed.

Definition closure (A : T -> Prop) := inter_s (fun B => close B /\ A <s B) id.

Notation "/ A" := (closure A).

Lemma closure_lc A B : close A -> B <s A -> /B <s A.
Proof. by move=> Hf HAB x; apply. Qed.

Lemma in_closure A : A <s /A. 
Proof. by move=> x Hx S [_ HS]; apply: HS. Qed.

Lemma close_closure A : close (/A).
Proof. by apply: inter_close => S []. Qed.

Add Morphism closure with signature (@eqset T) ==> (@eqset T) as closure_mor.
Proof.
move=> A B HAB; apply: double_i; apply: closure_lc; try exact: close_closure.
  by rewrite HAB; apply: in_closure.
by rewrite -HAB; apply: in_closure.
Qed.

Lemma closure_close_eq A : close A <-> /A <s A.
Proof.
split=> H; first exact: closure_lc.
have ->: A =s /A by apply: double_i => //; apply: in_closure.
exact: close_closure.
Qed.

Lemma closure_union A B : /(union A B) =s union (/A) (/B).
Proof.
split=>[Hx|].
  case: (classic ((/A) x)); first by left.
  case/not_all_ex_not=> S; move/not_imply_to_and=> [[H1 H2] H3].
  right=> U [HU1 HU2]; move: (Hx (union S U))=> Hf.
  have H: close (union S U) /\ union A B <s union S U. 
    split; first exact: union_close.
      exact: in_union.
  by case: (Hf H).
case=> H S [H1 H2]; apply: H; split=> //; apply: (include_trans _ H2).
  exact: union_left.
exact: union_right.
Qed.

Lemma closure_inter A B : /(inter A B) <s inter (/A) (/B).
Proof.
move=> x Hx. 
by split=>S [H1 H2]; apply: Hx; split=>//; apply: (include_trans _ H2)=>y [] //.
Qed.

Lemma clsr_cmpl_int A : /(A^c) =s (A°)^c .
Proof.
split=> [Hx|/not_ex_all_not Hx S [HS1 HS2]].
  apply: all_not_not_ex=> S.
  apply/andPC/not_and_to_imply=> [][H1 /compl_include H2].
  by apply: (Hx S^c); split=> //; rewrite /close compl_invo.
have/andPC/not_and_to_imply H := (Hx S^c).
apply/(compl_invo S)/H.
by split=> //; apply/compl_include; rewrite compl_invo.
Qed.

Lemma int_cmpl_clsr A : (A^c)° =s (/A)^c.
Proof.
split.
  case=> S [[HS1 HS2] HS3]; apply: ex_not_not_all.
  exists S^c; apply/not_imply_to_and.
  split; last exact/(compl_invo S).
   split=> //.  
     by rewrite /close  compl_invo.
   by apply/compl_include; rewrite compl_invo.
case/not_all_ex_not=> S /not_imply_to_and [[HS1 /compl_include HS2] HS3].
by exists S^c.
Qed.

Definition boundary (A : T -> Prop) := (/A) \ (A°).

Add Morphism boundary with signature (@eqset T) ==> (@eqset T) as boundary_mor.
Proof. by move=> A B HAB; rewrite /boundary HAB. Qed.

Lemma partition1 A x : x 'in A  -> x 'in A° \/ boundary A x. 
Proof.
move/in_closure=> H; case: (classic (boundary A x))=> HA; first by right.
by left; case/not_and_or: HA=> // Ho; apply/(compl_invo A°).
Qed.

Lemma partition2 A x : x 'in A^c \/ x 'in A° \/ boundary A x.
Proof.
case: (classic (x 'in A))=> HA; last by left.
by right; apply: partition1.
Qed.

Lemma boundary_eq_compl A : boundary A =s boundary A^c.
Proof.
by rewrite /boundary clsr_cmpl_int int_cmpl_clsr /without compl_invo inter_sym.
Qed.

Definition neighbourhood (V : T -> Prop) (x : T) := 
  exists A, [/\ x 'in A, open A & A <s V].

Definition neigh_open (V : T -> Prop) (x : T) := 
  neighbourhood V x /\ open V.

Lemma in_neigh V x : neighbourhood V x -> x 'in V.
Proof. by case=> S [H1 _ H2]; apply: H2. Qed.

Lemma neighbourhood_full : neighbourhood {full : T} =s {full : T}.
Proof. by split=>// _; exists {full : T}; split=>//; apply: full_in_open. Qed.

Lemma open_neigh A : 
  open A <-> (forall x, x 'in A -> (neighbourhood A x)).
Proof.
split=> [H x Hx|H]; first by exists A; split.
by apply/int_open_eq=> x Hx; case: (H x Hx)=> S [HS1 HS2]; exists S.
Qed.

Lemma int_m A : A° =s fun x => exists V, ((neighbourhood V x) /\ V <s A).
Proof.
split=> [[] S [[H1 H2] H3]|[] S [[V [HV HVo HVS]]] H].
  by do 2! (exists S; split=> //).
by exists V; do 2! (split=> //); apply: (include_trans HVS H).
Qed.

Lemma closure_m A : 
 (/A) =s 
 fun x => (forall V, (neighbourhood V x) -> exists y, y 'in (inter V A)).
Proof.
move=> x.
have HC : x 'in (/A)^c <->  (neighbourhood A^c x).
  split=> H.
    case: (not_all_ex_not _ _ H)=> F.
    move/not_imply_to_and=> [[H1 H2] H3].
    exists (F^c); split=>//.
    by move/compl_include: H2.
  case: H=> F [HF1 HF2 HF3].
  apply: ex_not_not_all; exists (F^c).
  apply/not_imply_to_and.
  split; last by move=> Hfin.
  split; rewrite /close ?compl_invo //. 
  by apply/compl_include; rewrite compl_invo.
split=> H S.
  case=> V [HV HV1 HV2]; case: (classic (exists y, y 'in (inter S A)))=> // Hy.
  suff: x 'in (/A)^c => //.
  apply/HC; exists V; split=> // y /HV2 Hy2.
  have Hab:= ((not_ex_all_not _ _ Hy) y).
  by case: (not_and_or _ _ Hab)=> //.
case: (classic ((/A) x))=> [|/HC Hab]; first by apply. 
by case: (H A^c Hab)=> y [].
Qed.

Definition separated := forall (x y : T), x <> y -> exists V1, exists V2,         
 [/\ (neighbourhood V1 x), (neighbourhood V2 y) &
     (inter V1 V2) <s {ø : T}].

Lemma separated_eq : separated <-> 
  (forall (x y : T), x <> y -> exists V1, exists V2,         
 [/\ (neigh_open V1 x), (neigh_open V2 y) &
     (inter V1 V2) <s {ø : T}]).
Proof.
split=> H x y Hxy; case: (H x y Hxy)=> V1; case=> V2 [H1 H2 H3].
  case: H1=> S [HS1 HS2 HS3]; case: H2=> U [HU1 HU2 HU3]; exists S; exists U.
  split; try split=> //.
      by exists S; split.
    by exists U; split.  
  by move=> z [/HS3 Hz1 /HU3 Hz2]; apply: H3. 
by exists V1; exists V2; case: H1; case: H2. 
Qed.

Lemma fin_inter : forall (s : seq (T -> Prop)), 
   (forall g, g \in s -> open g) -> open (inter_seq s).
Proof.
elim=> [H|A s HR H] /=; first exact: full_in_open. 
apply: inter_open; first by apply/H/mem_head.
by apply: HR=> g He; apply/H/mem_behead.
Qed.

Definition open_family Ti (f : family T Ti) := 
  forall i, i 'in (ind f) -> open (f i).

Lemma open_inter_fini (Ti : eqType) (f : finite_family T Ti) : 
  open_family f ->  open (inter_fam f).
Proof.
move=> Ho; rewrite (inter_fin_eq f); apply: fin_inter. 
move=> g; case H: (ind_seq f)=>[|t s]; rewrite /fam_fin_to_seq H //.
rewrite in_cons; case/orP=>[/eqP ->|].
  by apply: Ho; rewrite ind_seqP H;apply: mem_head.
rewrite -/(map _ _).
case/(nthP (f t))=> i; rewrite size_map=> Hi <-.
rewrite (nth_map t) //; apply: Ho.
by rewrite ind_seqP H; apply/mem_behead/mem_nth.
Qed.

Definition quasi_compact (K : T -> Prop) :=
   forall Ti (f : family T Ti),
   open_family f -> cover K f -> 
   exists g : finite_family T Ti, sub_family f g /\ cover K g.

Add Morphism quasi_compact with signature (@eqset T) ==> iff
  as quasi_compact_mor.
Proof.
move=> K1 K2 H12; split=> H Ti f Hof Hrf.
  have Hr : cover K1 f by rewrite /cover H12. 
  case: (H Ti f Hof Hr)=> g [Hg1 Hg2]; exists g.
  by split=> //; rewrite /cover -H12.
have Hr : cover K2 f by rewrite /cover -H12. 
case (H Ti f Hof Hr)=> g [Hg1 Hg2]; exists g.
by split=> //; rewrite /cover H12.
Qed.

Definition compact (K : T -> Prop) := separated /\ quasi_compact K.

Add Morphism compact with signature (@eqset T) ==> iff as compact_mor.
Proof. by move=> A B; rewrite /compact=> ->. Qed.

Lemma compct_close K : compact K -> close K.
Proof.
case; move/separated_eq=> SP HK; apply/int_open_eq.
rewrite int_m=> x Hx.
pose V := mkfamily K (fun i y : T => (forall Hi : K i,
             pi1 (cid (pi2 (cid (SP x i (neq_compl Hx Hi))))) y)).
have Hr : cover K V.
  move=> z Hz; exists z => /=; split=> //.
  move=> H2; apply: in_neigh.
  by case: (pi2 (cid (pi2 (cid (SP _ _ (neq_compl Hx H2))))))=> _ [H _ _].
have Ho : open_family V.
  move=> z Hz; apply/open_neigh => y Hy /=.
  exists (pi1 (cid (pi2 (cid (SP _ _ (neq_compl Hx Hz)))))).
  split; first exact: Hy.
  by case: (pi2 (cid (pi2 (cid (SP _ _ (neq_compl Hx Hz))))))=> _ [_ H _].
  move=> w Hw /= Hz2. 
  by have -> : Hz2 = Hz by apply: proof_irrelevance.  
case: (HK _ V Ho Hr)=> U [[H1 H2] H3].  
pose V' := mkfamily (ind U) (fun i y => (forall Hi : K i, 
                          pi1 (cid (SP x i (neq_compl Hx Hi))) y)).
have HfV': finite_set (ind V') by apply: finfam U.
pose fV' := Ff HfV'.
exists (inter_fam fV'); split.
  exists (inter_fam fV').
  split=> //.
    move=> y Hy /= Hi; apply: in_neigh.  
    by case: (pi2 (cid (pi2 (cid (SP _ _ (neq_compl Hx Hi))))))=> [][H _] _.
  apply: open_inter_fini.
  move=> z Hz; apply/open_neigh => y Hy /=.
  have Hy2 : K z by apply H1.
  exists (pi1 (cid (SP _ _ (neq_compl Hx Hy2)))).
  split; first exact: Hy.
  by case: (pi2 (cid (pi2 (cid (SP _ _ (neq_compl Hx Hy2))))))=> [][_ H] _.
  move=> w Hw /= Hz2. 
  by have -> : Hz2 = Hy2 by apply: proof_irrelevance. 
move=> y Hy Habs.
case: (H3 y Habs)=> z [Hz1 Hz2].
have Hz : K z by apply: H1.
have [_ _ Hv]:= ((pi2 (cid (pi2 (cid (SP _ _ (neq_compl Hx Hz))))))).
suff: False=> //.
apply: (Hv y); split.
  have Hbz : ind V' z by[].
  by apply: (inter_fam_in Hbz).
by move/H2: Hz2; apply. 
Qed.

Fixpoint ss_opt (T1 : Type) (s : seq (option T1)) := 
    match s with
     | [::]=> [::]
     | x :: s'=> match x with
                     |Some i => i :: ss_opt s'
                     |None => ss_opt s'
                   end
    end.

Lemma in_ss_opt : forall (T1 : eqType) (s : seq (option T1)) (x : T1),
   (Some x) \in s  <-> x \in ss_opt s. 
Proof.
move=> U; elim=> // t s HR x; case: t.
  move=> s0 /=; rewrite !in_cons; split; case/orP.
     -by move=> H ; apply/orP; left.
     -by move=> H; apply/orP; right; apply/HR.
     -by move/eqP=> ->; apply: mem_head.
     -by move=> H; apply/orP; right; apply/HR.
rewrite /= in_cons; split.
  -by case/orP=> H //; apply/HR.
  -by move=> H; apply/orP; right; apply/HR.
Qed.

Lemma close_in_comp K A : compact K -> close A -> A <s K -> compact A.
Proof.
move=> [HS HK] HA HKA; split=> // Ti V Ho Hr.
pose V' := mkfamily (oapp (ind V) true) (oapp V A^c).
have HrK : cover K V'. 
  move=> x HKx; case: (classic (x 'in A))=> HAx.
    by case: (Hr x HAx)=> y HBy; exists (Some y).
  by exists None.
have HoK: open_family V' by case => //=; apply: compl_f_o.
case: (HK _ V' HoK HrK)=> U [[H1 H2] H3].
pose W := mkfamily (fun i => ind U (Some i)) (fun i => U (Some i)).
have HfW: finite_set (ind W).
  pose s := ss_opt (ind_seq U).
  exists s => x /=; split=> H.
    by apply/in_ss_opt/ind_seqP.
  by apply/ind_seqP/in_ss_opt.
pose fW := Ff HfW.
exists fW; split=> [|x Hx].
  split=> x Hx.
    by have Hi: (Some x) 'in (ind V') by apply: H1.
  have -> : W x =s U (Some x) by [].
  by rewrite -H2.
have HKx : K x by apply: HKA.
case: (H3 x HKx); case.
  move=> t [Ht1 Ht2].
  by exists t.
by case=> HnA; move/H2=> Hn2; move: (Hn2 HnA).
Qed.

End TopologyTheory.

Existing Instance open_mor_Proper. 
Existing Instance interior_mor_Proper.
Existing Instance close_mor_Proper.
Existing Instance closure_mor_Proper.
Existing Instance boundary_mor_Proper.
Existing Instance quasi_compact_mor_Proper.
Existing Instance compact_mor_Proper.

Notation "A °" := (interior A) (at level 0, no associativity).
Notation "/ A" := (closure A).

Section Topology_sequence.

Variable (T : topologyType).

Definition Un_cvg (Un : nat -> T) (l : T) := 
  forall V, (neighbourhood V l) -> 
  exists N : nat, forall n, N <= n -> (Un n) 'in V.
Notation "Un >->> l" := (Un_cvg Un l) (at level 0, no associativity).

Definition cluster_point (Un : nat -> T) (l : T) := 
  forall V, (neighbourhood V l) -> 
  forall N : nat, exists n, N <= n /\ (Un n) 'in V.

Definition cluster_point2 (Un : nat -> T) (l : T) := 
  forall V, (neigh_open V l) -> 
  forall N : nat, exists n, N <= n /\ (Un n) 'in V.

Lemma cluster_point2_eq (Un : nat-> T) : 
  cluster_point Un =s cluster_point2 Un.
Proof.
move=> l; split=> H V.
  by move=> [H1 H2]; apply: H.
case=> S [HS HS1 HS2]=> N.
have HVS : neigh_open S l by split=> //; move/open_neigh: HS1; apply.
by case: (H S HVS N)=> n [H1 H2]; exists n; split=> //; apply: HS2. 
Qed.

Lemma cvg_const (a : T) : (fun _ => a) >->> a.
Proof.
move=> S HS.
exists 0%N=> n _.
exact: in_neigh.
Qed.

Lemma cvg_ext (a : T) (U V : nat -> T) :
  V =1 U -> U >->> a -> V >->> a.
Proof.
move=> HUV HU S HS.
have [N HN]:= HU S HS.
by exists N => n Hn; rewrite HUV; apply: HN.
Qed.

Lemma cvg_shift (a : T) (U : nat -> T) k :
   U >->> a <-> (fun i => U (i + k)%N) >->> a.
Proof.
split=> Hcvg S HS; have[N HN]:= Hcvg S HS.
  by exists (N - k)%N=> n Hn; apply: HN; rewrite addnC -leq_subLR.
exists (N + k)%N=> n Hn.
have/(leq_trans (leq_addl N k)) Hn2 := Hn.
by rewrite -[n](subnK Hn2); apply: HN; rewrite -(leq_add2r k) subnK.
Qed.

Lemma cvg_shiftS (U : nat -> T) l :
 (fun n => U n.+1) >->> l <-> U >->> l.
Proof.
have [H1 Hl]:= cvg_shift l U 1.
have ext k : U (k + 1)%N = U k.+1 by rewrite addn1.
split=> H; first by apply/Hl/(cvg_ext ext).
by apply/(cvg_ext (fsym ext))/H1.
Qed.

Definition fam_clstr (Un : nat -> T) (n : nat) (x : T) :=
  exists k, n <= k /\ x = Un k.

Definition cluster_point3 (Un : nat -> T) := 
    inter_fam (mkfamily {full : nat} (fun n x => (/(fam_clstr Un n)) x)). 

Lemma cluster_point3_eq Un : cluster_point Un =s cluster_point3 Un.
Proof.
apply: double_i => x Hx.
  move=> N _ ; apply/closure_m => V HV.
  case: (Hx V HV N)=> n [H1 H2] .
  by exists (Un n); split=> //; exists n.
rewrite /cluster_point2 /inter_fam /= in Hx.
move=> V HV N.
have t : True by [].
move/closure_m: (Hx N t)=> Hx2.
case: (Hx2 V HV)=> y [Hy1 [n [Hy3 Hy4]]].
by exists n; split=> //; rewrite -Hy4.
Qed.

Lemma lim_unicity Un l1 l2 : (separated T) ->
   Un >->> l1 -> Un >->> l2 -> l1 = l2.
Proof.
move=> HS H1 H2.
case: (classic (l1 = l2))=> // H.
case: (HS _ _ H)=> V1; case=> V2 [HV1 HV2 HV].
case: (H1 _ HV1)=> N1 HN1; case: (H2 _ HV2)=> N2 HN2.
pose N := maxn N1 N2.
have H1N : N1 <= N by rewrite leq_max; apply/orP; left.
have H2N : N2 <= N by rewrite leq_max; apply/orP; right.
have/HV Habs : (Un N) 'in (inter V1 V2).  
  split; first exact: HN1.
  exact: HN2.
done.
Qed.

Structure sub_seq : Type := mkss
   {
     phi :> nat -> nat;
     grow : forall m n, (m < n)%N -> (phi m < phi n)%N
   }.

Lemma grow_contra : forall (g : sub_seq) m n, g n <= g m -> n <= m.
Proof.
move=> g m n; rewrite leqNgt (leqNgt n m).
by apply: contra; apply: grow.
Qed.

Lemma grow_eq : forall g : nat ->nat, 
   (forall k, (g k < g k.+1)%N) <-> (forall m n, (m<n)%N -> (g m < g n)%N ).
Proof.
move=> g; split => H; last by move=> k; apply: (H k k.+1).
by move=> m n; elim: n m => // n HR m; rewrite ltnS leq_eqVlt => h; case/orP: h;
 [move/eqP => -> 
 | move=> h; apply: (ltn_trans (HR m h))].
Qed.

Lemma ss_min_id : forall (g : sub_seq) n, n <= g n.
Proof.
move=> g; elim=> // n HR.
apply: (leq_ltn_trans HR); clear HR; move: n; apply/grow_eq.
exact (grow g).
Qed.

Lemma cvg_ss_cvg : forall (Un : nat -> T) (g : sub_seq) l,
   Un >->> l -> (Un \o g) >->> l.
Proof.
move=> Un g l H V HV.
case: (H V HV)=> N1 HN1.
exists N1=> n Hn.
by apply: HN1; apply: (leq_trans Hn (ss_min_id g n)).
Qed.

Lemma ss_cvg (Un : nat -> T) l : 
  (exists g : sub_seq, (Un \o g) >->> l) -> l 'in (cluster_point Un).
Proof.
case=> g Hg V HV N.
case: (Hg V HV)=> N2 HN2.
exists (maxn (g N) (g N2)).
have : N <= (maxn (g N) (g N2)) /\ N2 <= (maxn (g N) (g N2)).
  by rewrite !leq_max; split; apply/orP; [left | right]; apply: ss_min_id.
case=> [HN HN22]; split=> //.
rewrite /maxn; case H: ((g N) < (g N2)); apply: HN2 => //.
by apply: (@grow_contra g); move/negbT: H; rewrite -leqNgt.
Qed.

Lemma Bolzano_Weierstrass : forall (Un : nat -> T) K , compact K ->
   (forall n, (Un n) 'in K) -> exists l, l 'in (cluster_point Un).
Proof.
move=> Un K [_ HK] HUn.
case: (classic (exists l : T, l 'in (cluster_point Un))) => //.
move/not_ex_all_not=> Habs; suff : False => //.
have H n : n 'in (cluster_point2 Un)^c by apply/cluster_point2_eq.
have H1 l : exists V,  ~((neigh_open V l) ->
              forall N : nat, exists n, N <= n /\ (Un n) 'in V).
  by apply: not_all_ex_not; apply: H.
have H2 l : exists V, (neigh_open V l) /\ 
             ~(forall N : nat, exists n, N <= n /\ (Un n) 'in V).
  by case: (H1 l)=> V HV; exists V; apply/not_imply_to_and.
have H3 l : exists V, (neigh_open V l) /\ 
              exists N, forall n, ~ (N <= n) \/ (Un n) 'in V^c.
  case: (H2 l)=> V [HV1 HV2].
  exists V; split=> //; case/not_all_ex_not: HV2=> N. 
  by move/not_ex_all_not=> HN; exists N=> n; apply: not_and_or; apply: HN.
clear H Habs H1 H2.
pose f := (pi1 (cid (H3 _))).
pose W := mkfamily K f.
have Hr : cover K W.
  move=> x Hx /=; exists x; split => //; apply: in_neigh.
  by apply: (proj1 (proj1 (pi2 (cid (H3 x))))).
have Ho :  open_family W.
  by move=> x Hx; apply: (proj2 (proj1 (pi2 (cid (H3 x))))).
case : (HK _ W Ho Hr)=> U [[HU1 HU2] HU3].
have:= ind_seqP U.
move: (ind_seq U); case=> [HU4 | t s HU4].
  have Hkf : forall x, K x -> False.
    by move=> x Hx; case: (HU3 x Hx)=> y [Hbs _]; move/HU4: Hbs.
  by move: (Hkf (Un 0%N) (HUn 0%N)).
pose P := fun N => forall n, N <= n -> ~ U _ (Un n).
pose g := (pi1 (cid (proj2 (pi2 (cid (H3 _)))))).
have Habs : forall x, x \in t :: s -> P x (g x).
  move=> x; move/HU4=> Hx n Hn.
  apply/(HU2 x Hx).
  move/or_to_imply: ((pi2 (cid (proj2 (pi2 (cid (H3 x)))))) n).  
  by apply.
pose m := \max_(x <- t :: s) g x.
  rewrite (big_nth t) big_mkord in m.
have: { i :  'I_(size (t :: s)) | m = g (nth t (t :: s) i)}.
   by apply: eq_bigmax; rewrite card_ord.
case=> i Hi.
pose x := (nth t (t :: s) i).
pose n := g x.
have Hm : forall j :'I_(size (t :: s)), g (nth t (t :: s) j) <= n.
  by move=> j; rewrite /n -Hi; apply: (leq_bigmax j).
have H : forall y, y \in t :: s -> index y (t :: s) < size (t :: s).
   by move=> y Hy; rewrite index_mem. 
have HP : forall y, y \in t :: s ->
  P y (g (nth t (t :: s) (index y (t :: s)))).
  by move=> y Hy; rewrite nth_index //; apply: Habs.
have Hfin : forall x, ind U x -> ~ U x (Un n).
  move=> y; move/HU4=> Hy; apply: (HP y Hy); rewrite insE -index_mem in Hy.  
  exact (Hm (Ordinal Hy)).
by case: (HU3 (Un n) (HUn n))=> y [Hy1 Hy2]; apply: (Hfin y Hy1).
Qed.

Lemma close_lim A : close A -> 
 (forall (Un : nat -> T) l, (forall n, (Un n) 'in A) -> Un >->> l -> l 'in A).
Proof.
move/closure_close_eq=> H Un l HA Hcv; apply/H/closure_m=> V HV.
by case: (Hcv V HV)=> n Hn; exists (Un n); split=> //; apply: (Hn n).
Qed.

End Topology_sequence.

Notation "Un >->> l" := (Un_cvg Un l) (at level 0, no associativity).

Section Topology_function.

Variables T1 T2 : topologyType.

Definition continue_pt (f : T1 -> T2) (x : T1) :=
  forall V', (neighbourhood V' (f x)) ->
  exists V, (neighbourhood V x) /\ f^(V) <s V'.

Definition continue (f : T1 -> T2) := forall x, continue_pt f x.

Lemma continue_open : forall (f : T1 -> T2),
  (continue f <-> forall U, open U -> open f@^-1(U)).
Proof.
move=> f; split => H.
  move=> U HU.
  apply/int_open_eq=> x Hx; case: (H x U). 
    by exists U; split; case: Hx => y [Hy1 Hy2] //; rewrite Hy2 in Hy1.
  move=> S [] [A [HA1 HA2 HA3]] HSU.
  exists A; split=> //; split=> // y Hy.
  exists (f y); split=> //; apply: HSU.
  by exists y; split=> //; apply: HA3.
move=> x V' [] A [HA1 HA2 HA3]. 
exists f@^-1(A); split.
  exists f@^-1(A); split=> //; first by exists (f x).
  exact: H.
by apply: (include_trans _ HA3); apply: f_finv_in.
Qed.

Lemma continue_close : forall (f : T1 -> T2),
  (continue f <-> forall U, close U -> close f@^-1(U)).
Proof.
move=> f; rewrite /close; split=> H.
  by move=> U; rewrite -compl_finv; case: (continue_open f)=> Hf _; apply: Hf.
have H2 U: open (U)^c -> open (f) @^-1(U^c).
  by rewrite compl_finv; apply: H.
by apply/continue_open=> U; rewrite -(compl_invo U); apply: H2.
Qed.

Lemma continue_const (x : T2) : continue (fun (_ : T1) => x).
Proof.
move=> y A Hx.
exists {full : T1}; split; first by apply/neighbourhood_full.
by move=> z [w [_ ->]]; apply: in_neigh. 
Qed.

Lemma continue_ext (f g : T1 -> T2) : f =1 g -> continue f -> continue g.
Proof.
move=> Hfg Hf x r; rewrite -Hfg => H.
have [V [Hx HV]] := Hf x r H.
exists V; split=> // y [z Hz].
by apply: HV; exists z; rewrite Hfg.
Qed.

Lemma cont_cmpct f K : separated T2 -> continue f -> compact K -> compact (f^(K)).
Proof.
move=> HS Hf [H1 H2]; split=> // Ti U Ho Hr.
pose V := mkfamily (ind U) (fun i : Ti => (f) @^-1( U i)).
have HoV : open_family V.
   by move=> x Hx; case:(continue_open f)=> H _; apply: H=> //; apply: Ho.
have HrV : cover K V.
   move=> x Hx; case: (Hr (f x)); first by exists x.
   by move=> y [Hy1 Hy2]; exists y; split=> //; exists (f x).
case: (H2 _ V HoV HrV)=> W [[HW1 HW2] HW3].
pose V' := mkfamily (ind W) (fun i x => ind W i /\ U i x).
have HfV': finite_set (ind V') by apply: (finfam W).
pose fV' := Ff HfV'.
exists fV' ; split.
  split=> //.
  by move=> i Hi x; split=> //= [][_ Hx]. 
move=> x; case=> y [Hy1 ->]; case: (HW3 y Hy1)=> z [Hz1 Hz2].
exists z; split => //.
by move/(HW2 _ Hz1): Hz2=> /=; case=> w [H <-].
Qed.

Lemma continue_cvg f : continue f -> 
  (forall (Un : nat -> T1) l, Un >->> l -> (f \o Un) >->> (f l)).
Proof.
move=> Hf Un l H V HV; case: (Hf l V HV)=> V' [H1 H2]; case: (H V' H1)=> N HN.
by exists N=> n Hn; apply: H2; exists (Un n); split=>//; apply: HN.
Qed.

End Topology_function.    
 
Section Topology_function2.
  
Variables T1 T2 T3 : topologyType.

Lemma continue_id : continue (@id T1).
Proof. by move=> x A Hx; exists A; split=> // y [z [zinA ->]]. Qed.

Lemma continue_comp : forall (f : T1 -> T2) (g : T3 -> T1),
 continue f -> continue g -> continue (f \o g). 
Proof.
move=> f g; move/continue_open=> Hf; move/continue_open=> Hg.
by apply/continue_open=> U HU; rewrite comp_inv; apply: Hg; apply: Hf.
Qed.

End Topology_function2.   

Module EspMet.

Open Scope ring_scope.

Structure mixin_of (R : numDomainType) (T : Type) := Mixin {
    dist : T -> T -> R;
    _ : forall x y,  0 <= dist x y ;
    _ : forall x y, dist x y = dist y x;
    _ : forall x y, reflect (x=y) (dist x y == 0);
    _ : forall x y z, dist x y <= dist x z + dist z y
   }. 

Section ClassDef.

Variable R : numDomainType.

Notation class_of := (mixin_of R) (only parsing).

Structure type (phR : phant R) := Pack {sort; _ : class_of sort; _ : Type}.
Local Coercion sort : type >-> Sortclass.

Variable (phR : phant R) (T : Type) (cT : type phR).
Definition class := let: Pack _ c _ as cT' := cT return class_of cT' in c.
Definition pack c := @Pack phR T c T.
Definition clone := fun c & cT -> T & phant_id (pack c) cT => pack c.

End ClassDef.

Module Exports.

Coercion sort : type >-> Sortclass.
Notation espMetType R := (type (Phant R)).
Notation EspMetMixin := Mixin.
Notation EspMetType R T m := (@pack _ (Phant R) T m).
Notation "[ 'espMetMixin' R 'of' T ]" := (class _ : mixin_of R T)
  (at level 0, format "[ 'espMetMixin' R 'of' T ]") : form_scope.
Notation "[ 'espMetType' R 'of' T 'for' C ]" := (@clone _ (Phant R) T C _ idfun id)
  (at level 0, format "[ 'espMetType' R 'of' T 'for' C ]") : form_scope.
Notation "[ 'espMetType' R 'of' T ]" := (@clone _ (Phant R) T _ _ id id)
  (at level 0, format "[ 'espMetType' R 'of' T ]") : form_scope.

End Exports.
End EspMet.
Export EspMet.Exports.

Section positive.

Variable R : numDomainType.
Open Scope ring_scope.

Import GRing.Theory.
Import Num.Theory.
Import Num.Def.
 
Lemma inv_INR_pos k : (0 : R) < k.+1%:R^-1.
Proof. by rewrite invr_gt0 ltr0n. Qed.

Structure posDomain := PosD { sort :>  R ; _ : 0 < sort }. 

Definition posDomain_of of phant R := posDomain.

Canonical Structure natmulV_pos k := PosD (inv_INR_pos k). 

Fact lt_pos (r : posDomain) : 0 < r. Proof. by case: r. Qed.

Lemma ler_minl_pos (x y : posDomain) : minr x y <= x.
Proof.
rewrite /minr.
case: (real_lerP (gtr0_real (lt_pos x)) (gtr0_real (lt_pos y)))=> //.
exact: ltrW.
Qed.

Lemma ler_minr_pos (x y : posDomain) : minr x y <= y.
Proof.
rewrite /minr.
by case: (real_lerP (gtr0_real (lt_pos x)) (gtr0_real (lt_pos y))).
Qed.

Lemma minr_gt0 (x y : R) : 0 < x -> 0 < y -> 0 < minr x y.
Proof. by move=> xlt0 ylt0; rewrite /minr; case: (x <= y). Qed.

Lemma minr_pos (x y : posDomain) : 0 < minr x y.
Proof. by rewrite  minr_gt0 // !lt_pos. Qed.

Canonical Structure posD_minr x y := PosD (minr_pos x y).

Lemma inv_pos (r : posDomain) : 0 < (sort r)^-1.
Proof. by rewrite invr_gt0 lt_pos. Qed.

Canonical Structure posD_inv r := PosD (inv_pos r).

Lemma ltn_ltrV (F : numFieldType) k n : 
 (k < n)%N = (n.+1%:R^-1 < k.+1%:R^-1 :> F).
Proof.
rewrite /= -(ltr_pmul2l (ltr0Sn _ k)) -(ltr_pmul2r (ltr0Sn _ n)).
by rewrite -mulrA mulrV ?mulVr ?unitf_gt0 ?ltr0Sn // mulr1 mul1r ltr_nat.
Qed.

Lemma lt_eps_le (T : realDomainType) (a b : T) : 
  (forall eps, 0 < eps -> a < b + eps) -> a <= b.
Proof.
move=> H.
rewrite lerNgt -subr_gt0; apply/negP=>/(H (a - b)).
by rewrite addrC addrNK ltrr.
Qed.

Lemma real_lt_eps_le (a b : R) : a \is Num.real -> 
  (forall eps, 0 < eps -> a < b + eps) -> a <= b.
Proof.
move=> ra H.
have/ltrW/ler_real/(@sym_eq _ _ _):= H _ ltr01.
rewrite ra=>/(realB (x:=1)) Hb.
have:= Hb (real1 R); rewrite opprD [-_-_]addrC addNKr realN=> rb.
rewrite real_lerNgt // -subr_gt0; apply/negP=>/(H (a - b)).
by rewrite addrC addrNK ltrr.
Qed.

End positive.

Notation "{ 'posD' T }" := (posDomain_of (Phant T)).

Hint Resolve lt_pos.

Section positive_archi.

Variable R : archiDomainType.
Open Scope ring_scope.

Import Num.Theory.
Import Num.Def.

Lemma ex_sub_proof (x : {posD R}) : exists n : nat, x <= n%:R.
Proof. by exists (boundD x); rewrite ltrW // archiD_boundP // ltrW. Qed.

Definition pceil (x : {posD R}) := ex_minn (ex_sub_proof x). 

Lemma pceilE (x : {posD R}) : (pceil x).-1%:R < x <= (pceil x)%:R.
Proof.
rewrite /pceil; case: ex_minnP=> [][] =>[-> _ |m -> /(_ m)/contra].
  by rewrite lt_pos.
by  rewrite -real_ltrNge ?realn ?gtr0_real // ltnn => ->.
Qed.

Lemma leq_pceil (x : {posD R}) n : n%:R <= x -> (n <= (pceil x))%N. 
Proof. 
move=> lenx.
have /andP[_ lexc]:= pceilE x.
have := ler_trans lenx lexc.
by rewrite ler_nat.
Qed.

Lemma ex_up_proof (x : {posD R}) : exists n : nat, n%:R <= x.
Proof. 
exists (pceil x).-1; rewrite ltrW //. 
by have /andP[-> _]:= pceilE x.
Qed.

Definition pfloor (x : {posD R}) := ex_maxn (ex_up_proof x) (@leq_pceil x).

Lemma pfloorE x : (pfloor x)%:R <= x < (pfloor x).+1%:R.
Proof.
rewrite /pfloor; case: ex_maxnP=> i -> /(_ i.+1)/contra.
by rewrite -real_ltrNge ?realn ?gtr0_real // ltnn => ->.
Qed.

End positive_archi.

Section EspMet_Struct.

Variable (R : numDomainType). 
Variable E : (espMetType R).

Import Num.Theory.
Import Num.Def.

Open Scope ring_scope.

Notation d := (EspMet.dist (EspMet.class E)).

Lemma dist_pos :  forall x y : E,  0 <= d x y.
Proof. by case: E => ? []. Qed.

Lemma dist_sym : forall x y : E, d x y = d y x.
Proof. by case: E=> ? []. Qed.

Lemma dist_refl :  forall x y : E, reflect (x = y) (d x y == 0).
Proof. by case: E=> ? []. Qed.

Lemma dist_tri :  forall x y z : E, d x y <= d x z + d z y.
Proof. by case: E=> ? []. Qed.
 
Definition eqmet (x y : E) : bool := (d x y) == 0.

Lemma eqmetP : Equality.axiom eqmet.
Proof.
by move=> x y; apply/(iffP idP)=> /dist_refl H.
Qed.

Canonical Structure EspMet_eqMixin := Eval hnf in EqMixin eqmetP.
Canonical Structure EspMet_eqType := Eval hnf in EqType E EspMet_eqMixin.

Definition open_ball (x : E) (r : {posD R}) (y : E) : Prop := d x y < r.
Notation "''B' ( x , r )" := (open_ball x r)
  (at level 8, format "''B' ( x ,  r )").

Definition open_met (A : E -> Prop) := 
 forall x, x 'in A -> exists r, 'B(x,r) <s A.

Lemma empty_in_open_met : open_met {ø : E}.
Proof. by []. Qed.

Lemma full_in_open_met : open_met {full : E}.
Proof. by move=> x _; exists (PosD ltr01). Qed.

Lemma union_open_met : forall (S : (E -> Prop) -> Prop), S <s open_met ->
  open_met (union_s S id).
Proof.
move=> S HS x; case=> A [HA1 HA2].
case: (HS A HA1 x HA2)=> r Hr.
exists r; apply: (include_trans Hr).
by apply: (@in_unions _ _ id).
Qed.

Lemma inter_open_met : forall A B, open_met A -> open_met B ->
   open_met (inter A B).
Proof.
move=> A B HA HB x [Hx1 Hx2].
case: (HA x Hx1)=> r1 Hr1; case: (HB x Hx2)=> r2 Hr2.
exists (PosD (minr_gt0 (lt_pos r1) (lt_pos r2))).
move=> y Hy; split.
  by apply/Hr1/(ltr_le_trans Hy)/ler_minl_pos.
by apply/Hr2/(ltr_le_trans Hy)/ler_minr_pos.
Qed.

Lemma ext_open_met : forall A B, A =s B -> (open_met A <-> open_met B).
Proof.
move=> A B H; split=> Ho x Hx.
  elim: (Ho x ((right_i H) x Hx))=> r Hr; exists r.
  exact (include_trans Hr (left_i H)).
elim: (Ho x ((left_i H) x Hx))=> r Hr; exists r.
exact (include_trans Hr (right_i H)).
Qed.

Definition EspMet_TopMixin := Eval hnf in TopologyMixin 
  empty_in_open_met full_in_open_met union_open_met inter_open_met ext_open_met.
Definition EspMet_TopType := Eval hnf in
   TopologyType E EspMet_TopMixin.

Canonical EspMet_TopMixin.
Canonical EspMet_TopType.

End EspMet_Struct.

Notation d := (EspMet.dist (EspMet.class _)).
Notation "''B' ( x , r )" := (open_ball x r)
  (at level 8, format "''B' ( x ,  r )").

Section EspMet_Theory.

Section numDomainTypeT.

Variable R : numDomainType.
Variable E : (espMetType R).

Notation d := (EspMet.dist (EspMet.class E)).

Import GRing.Theory.
Import Num.Theory.
Import Num.Def.

Open Scope ring_scope.

Lemma dist_nul x : d x x = 0.
Proof. exact/eqP/dist_refl. Qed.

Lemma bxrx (x : E) r : x 'in 'B(x,r).
Proof. by rewrite /open_ball insE dist_nul. Qed.

Lemma ball_open : forall (x :  E) r, open 'B(x,r).
Proof.
move=> x r a H.
have p : 0 < (sort r) - d x a by rewrite subr_gt0.
pose s:= PosD p.
exists s => y h.
have t : (sort r) = (d x a + (sort s)) by rewrite addrC addrNK.
apply: (@ler_lt_trans _ (d x a + d a y)); first exact: dist_tri.
by rewrite t ltr_add2l. 
Qed.

Lemma ball_neigh : forall (x : E) r, neighbourhood 'B(x,r) x.
Proof.
move=> x r; exists 'B(x,r).
split=> //; first by apply: bxrx.
exact: ball_open.
Qed.
  
Lemma interior_met : forall (A : E -> Prop), 
   A° =s (fun x=>exists r, 'B(x,r) <s A). 
Proof.
move=> A; apply: double_i=> x [] => [S [[HS1 HS2] HS3]|r Hr].
  case: (HS1 x HS3)=> r Hr; exists r.
  exact: (include_trans Hr HS2).
exists 'B(x,r); split; last exact: bxrx.
by split => //; apply: ball_open.
Qed.

Lemma closure_met : forall (A : E -> Prop) ,
  /A =s (fun x=> forall r, exists y, y 'in (inter 'B(x,r) A)).
Proof.
move=> A; apply: double_i=> x.
  by move/closure_m=> Hx r; apply: (Hx 'B(x,r)); apply: ball_neigh.
move=> Hr; apply/closure_m => V; case=> B [HB1 HB2 HB3].
case: (HB2 x HB1)=> r HrB; case: (Hr r)=> y [Hy1 Hy2].
exists y; split=> //.
by apply/HB3/HrB.
Qed.

Definition bounded (A : E -> Prop) :=
  exists r, forall x y, x 'in A -> y 'in A -> y 'in 'B(x,r).

Definition bounded2 (A : E -> Prop) : Prop := 
   exists r, exists x, A <s 'B(x,r). 

Lemma born_eq : forall A, bounded2 A -> bounded A.
Proof.
move=> A; case=> r; case=> x H.
have Hp : 0 < (sort r) + (sort r) by rewrite addr_gt0.
exists (PosD Hp)=> y z Hy Hz. 
have Hd : d y x + d x z < (sort r) + (sort r).
  by rewrite ltr_add // ?(dist_sym y) H. 
by apply: (ler_lt_trans _ Hd); apply: dist_tri.
Qed.

Lemma pos : forall (x y : E), x<>y -> 0 < d x y.
Proof.    
move=> x y Hxy.
rewrite ltr_def dist_pos andbT. 
exact/dist_refl.
Qed.

Lemma pos_2 : forall (x y : E), x <> y -> 0 < (d x y)/2%:R.
Proof.
move=> x y Hxy.
by rewrite mulr_gt0 // ?pos // invr_gt0 ltr0n.
Qed.

Lemma comp_born : forall (K : E -> Prop), compact K -> bounded K.
Proof.
move=> K [HS H].
pose V := mkfamily K (fun x y => y 'in 'B(x,(PosD ltr01))). 
have Hr : cover K V.
  by move=> x HKx; exists x; split=> //; apply: bxrx.
have Ho : open_family V by move=> x Hx; apply: ball_open.
case: (H _ V Ho Hr)=> g [[Hg1 Hg2] Hg3].
have:= ind_seqP g.
case He: (ind_seq g)=> [|t s0].
  move=> Hg4; exists (PosD ltr01)=> x y HKx.
  case: (Hg3 x HKx)=> z [Habs _].
  by move/Hg4: Habs.
move=> Hs0 .
pose r:= \maxr_(i <- (ind_seq g)) d t i.
have [x /andP [_ /eqP Hx]]: {x |  (x \in (ind_seq g)) &&  (d t x == r)}.
  apply: bigmax_eq=>[i|]; first by rewrite dist_pos.
  by rewrite He.
have Hp1 : 0 < r + 1 by rewrite -[0]addr0 ler_lt_add // ?ltr01 // -Hx dist_pos.
have Hp : 0 < (r +1) + (r + 1) by rewrite addr_gt0. 
exists (PosD Hp)=> y z Hy Hz.
have Hi w : K w -> exists v, v \in (ind_seq g) /\ w 'in 'B(v,(PosD ltr01)).
  move=> Hw; case: (Hg3 w Hw) => a [Ha1 Ha2].
  exists a; split; first by rewrite He; apply/Hs0.
  by apply/Hg2.
case: (Hi y Hy)=> v [Hiv Hv].
case: (Hi z Hz)=> w [Hiw Hw].
apply: (ler_lt_trans (dist_tri _ _ t)).
rewrite ler_lt_add // ?(dist_sym y t) ?ltrW //.
  rewrite (ler_lt_trans (dist_tri _ _ v)) // ler_lt_add // bigmax_ger //.
  by move=> i; rewrite dist_pos.
rewrite (ler_lt_trans (dist_tri _ _ w)) // ler_lt_add // bigmax_ger //.
by move=> i; rewrite dist_pos.
Qed.

End numDomainTypeT.

Section numFieldTypeT.

Variable R : numFieldType.
Variable E : (espMetType R).

Notation d := (EspMet.dist (EspMet.class E)).

Import GRing.Theory.
Import Num.Theory.
Import Num.Def.

Open Scope ring_scope.

Lemma separated_esp_met : separated (EspMet_TopType E).
Proof.
move=> x y Hxy.
exists 'B(x,(PosD (pos_2 Hxy))).
exists 'B(y,(PosD(pos_2 Hxy))).
split; try by apply: ball_neigh.
move=> z [Hz1 Hz2].
have H : d x y = 2%:R * (d x y / 2%:R).
  by rewrite mulrA mulrC mulKr ?unitf_gt0 ?ltr0n.
have: d x z + d y z < d x y.
  by rewrite H mulrDl !mul1r ltr_add.
by rewrite real_ltrNge ?realD ?ger0_real ?dist_pos // (dist_sym y z) dist_tri.
Qed.

End numFieldTypeT.
 
Section archiFieldTypeT.

Variable R : archiDomainType.
Variable E : (espMetType R).

Notation d := (EspMet.dist (EspMet.class E)).

Import GRing.Theory.
Import Num.Theory.
Import Num.Def.

Open Scope ring_scope.

Lemma cont_to_discr (r : {posD R}) :
   (PosD (inv_INR_pos _ (pfloor (posD_inv r)).+1) < r).
Proof.
have /andP [_ H2]:= pfloorE (posD_inv r).
set n := pfloor _.
have Hlt0 : (0 : R) < n.+2%:R by rewrite ltr0Sn.
rewrite /= -(ltr_pmul2r Hlt0) mulVr ?unitfE ?pnatr_eq0 //.
by rewrite -ltr_pdivr_mull // mulr1 (ltr_trans H2) // ltr_nat.
Qed.

Lemma int_nat (A :  E -> Prop) :
A° =s (fun x=> exists n : nat, 'B(x,(PosD (inv_INR_pos _ n))) <s A).
Proof.
rewrite interior_met=> x; split; elim=> r Hr.
  exists (pfloor (posD_inv r)).+1.
  apply: (include_trans _ Hr)=> y Hy; apply: (ltr_trans Hy).
  exact: (cont_to_discr r).
by exists (PosD (inv_INR_pos _ r)).
Qed.

Lemma closure_nat : forall (A : E -> Prop), /A =s
(fun x=> forall n, exists y, y 'in (inter 'B(x,(PosD (inv_INR_pos _ n))) A)).
Proof.
move=> A; rewrite closure_met; split => H r.
  by elim: (H (PosD (inv_INR_pos _ r)))=> y Hy; exists y.
elim: (H (pfloor (posD_inv r)).+1)=> y [Hy1 Hy2]; exists y.
by split=> //; apply/(ltr_trans Hy1)/(cont_to_discr r).
Qed.

Lemma born_nat A : bounded A <->
  exists n, forall (x y : E), x 'in A -> y 'in A -> 
  y 'in 'B(x,(PosD (ltr0Sn _ n))).
Proof.
split; case=> r H.
  exists (pfloor r)=> x y Hx Hy.
  apply: (ltr_trans (H _ _ Hx Hy))=> /=.
  by have/andP [_] := pfloorE r.
by exists (PosD (ltr0Sn _ r)).
Qed.

End archiFieldTypeT.

End EspMet_Theory.

Section EspMet_sequence.

Section numDomainTypeTop.

Import GRing.Theory.
Import Num.Theory.
Import Num.Def.
Open Scope ring_scope.

Variable R : numDomainType.
Variable E : (espMetType R).

Definition Un_cvg_met (Un : nat -> E) (l : E) :=
  forall (eps : {posD R}), exists N : nat, 
  forall n, (N <= n)%N -> (Un n) 'in 'B(l,eps). 

Lemma Un_cvgMT : forall (Un : nat -> E) l, Un_cvg_met Un l <-> 
  Un >->> l.
Proof.
move=> Un l; split => H; [move => V  [S [HS1 HS2 HS3]] | move=> e].
  case: (HS2 l) => // e hB; case: (H e) => N Hf; exists N => n Hn; apply: HS3.
  exact ((hB (Un n)) (Hf n Hn)).
exact (H _ (ball_neigh l e)).
Qed.

End numDomainTypeTop.

Section archiDomainTypeTop.

Import GRing.Theory.
Import Num.Theory.
Import Num.Def.
Open Scope ring_scope.

Variable R : archiDomainType.
Variable E : (espMetType R).

Lemma Un_cvg3 : forall (Un : nat -> E) l, Un >->> l <->
  (forall (k : nat), exists N : nat, forall n, (N <= n)%N ->
  (Un n) 'in 'B(l,(PosD (inv_INR_pos _ k)))).
Proof.
move=> Un l; split.
  by move/Un_cvgMT=> H k; case (H (PosD (inv_INR_pos _ k)))=> N H2; exists N.
move=> H; apply/Un_cvgMT=> e; elim (H (pfloor (posD_inv e)).+1) => N H2.
exists N => n h; move: (H2 n h) => H1; apply: (ltr_trans H1).
exact: (cont_to_discr e).
Qed.

Lemma ex_ss_cvg : forall (Un : nat -> E),
  (exists x, exists g: sub_seq, forall (k:nat),
  (Un (g k)) 'in 'B(x,(PosD (inv_INR_pos _ k)))) ->
  (exists g : sub_seq, exists x,  (Un \o g) >->> x).
Proof.
move=> Un; case => x; case => g H; exists g; exists x.
apply/Un_cvg3 => k; exists k.+1 => n h.
by apply: (ltr_trans (H n)); rewrite -ltn_ltrV.
Qed.
 
Lemma ex_ss_cvg2 : forall (Un : nat -> E) l, 
 (forall N: nat, forall (k:nat), exists n: nat, 
 (N <= n)%N /\  (Un n.+1) 'in 'B(l,(PosD (inv_INR_pos _ k)))) -> 
 (exists g: sub_seq, (Un \o g) >->> l). 
Proof.
move=> Un l H.
have: exists G : nat -> (nat -> nat), forall N: nat, 
      forall k : nat , (N <= (G N) k)%N /\ 
      (Un ((G N) k).+1) 'in 'B(l,(PosD (inv_INR_pos _ k))) 
    by apply: (functional_choice (fun (N: nat) (f: nat -> nat) => forall k:nat,
       (N <= f k)%N /\ (Un (f k).+1) 'in 'B(l,(PosD (inv_INR_pos _ k))))) => N; 
       apply: (functional_choice (fun (k:nat) (n:nat) => 
        (N <= n)%N /\  (Un n.+1) 'in 'B(l,(PosD (inv_INR_pos _ k))))).
elim => G H1.
pose g := fix t m : nat := if m is m'.+1 then (G (t m') m').+1 else G 0%N 0%N.
have H2: forall m n, (m < n)%N -> (g m < g n)%N.
  by apply/grow_eq=> k; case: (H1 (g k) k).
exists (mkss H2) => /=; apply/Un_cvg3 => k; exists k.+1 => n h.
set r:= (PosD _).
rewrite -(ltn_predK h). 
case: (H1 (g n.-1) n.-1)=> _ H3.
apply/(ltr_le_trans H3).
rewrite ler_eqVlt -ltn_ltrV (inj_eq invr_inj) eqr_nat eqSS.
by rewrite  eq_sym -leq_eqVlt -ltnS (ltn_predK h).
Qed.

Lemma clstr_ss_cvg : forall (Un : nat -> E) l, l 'in (cluster_point Un) ->
  (exists g : sub_seq, (Un \o g) >->> l).
Proof.
move=> Un l H; apply: ex_ss_cvg2=> N k.
have HV : neighbourhood ('B(l,(PosD (inv_INR_pos _ k)))) l.
  by apply: ball_neigh.
case: (H _ HV N.+1)=> n [Hn H2].
exists n.-1; rewrite (ltn_predK Hn); split=> //.
by rewrite -ltnS (ltn_predK Hn).
Qed.


Lemma lim_close : forall (A : E -> Prop),
  (forall (Un : nat -> E) l, (forall n, A (Un n)) -> Un >->> l -> A l)
  -> close A.
Proof.
move=> A H; apply/closure_close_eq=> x.
move/closure_nat => Hx.
have: exists Un : nat -> E, forall n,
          inter ('B(x,(PosD (inv_INR_pos _ n)))) A (Un n)
        by apply: (functional_choice (fun (n : nat) (y : E) =>
         inter ('B(x,(PosD (inv_INR_pos _ n)))) A y)).
elim=> Un Hun.
apply: (H Un x); first by move=> n; move: (Hun n)=> [_ Ha].
apply/Un_cvg3=> k; exists k.+1 => n Hn.
have [Hu _]:= (Hun n).
by apply: (ltr_trans Hu); rewrite -ltn_ltrV.
Qed.

End archiDomainTypeTop.
End EspMet_sequence. 

Section EspMet_function.

Import Num.Theory.
Import Num.Def.

Open Scope ring_scope.
Variable R R': numDomainType.
Variables E : (espMetType R).
Variables E': (espMetType R').

Definition continue_pt_met (f : E -> E') (x : E) : Prop := 
  forall (r : {posD R'}), exists r' : {posD R}, forall (y : E),
  (y 'in 'B(x,r')) -> (f y) 'in 'B(f x,r).

Definition continue_met (f : E -> E') : Prop := forall x, continue_pt_met f x.

Lemma continue_ptMT : forall (f : E -> E') (x : E), 
  continue_pt_met f x <-> continue_pt f x.
Proof.
move=> f x; split=> H.  
  move=> V [S [H1 H2 H3]].
  case: (H2 (f x))=> // r Hr.
  case: (H r)=> r2 Hr2.
  exists 'B(x,r2).
  split; first by apply: ball_neigh.
  move=> y [z [Hz1 Hz2]].
  by apply: H3; apply: Hr; rewrite Hz2; apply: Hr2.  
move=> r.
case: (H _ (ball_neigh (f x) r))=> V [[S [H1 H2 H3]] H4].
case: (H2 x)=> // r2 Hr2.
exists r2=> y Hy.
apply: H4; exists y; split=> //.
by apply: H3; apply: Hr2.
Qed.

Lemma continueMT : forall (f : E -> E'), continue_met f <-> continue f.
Proof.
by move=> f; split=> H x; apply/continue_ptMT.
Qed.

End EspMet_function.

Section nummetrie.
Variable T : numDomainType.

Open Scope ring_scope.

Definition dist_real (x y : T) := `|x - y|.

Lemma dist_real_ge0 x y : 0 <= dist_real x y.
Proof. exact: normr_ge0. Qed.

Lemma dist_real_sym x y : dist_real x y = dist_real y x.
Proof. exact: distrC. Qed. 

Lemma dist_real_refl x y : reflect (x = y) (dist_real x y == 0).
Proof. 
rewrite /dist_real normr_eq0 subr_eq0.
exact: eqP.
Qed.

Lemma dist_real_tri x y z : dist_real x y <= dist_real x z + dist_real z y.
Proof. by rewrite /dist_real -{1}(addrNK z x) -addrA ler_norm_add. Qed.

Canonical Structure numEspMetMixin :=
  EspMetMixin dist_real_ge0 dist_real_sym dist_real_refl dist_real_tri.
Canonical Structure numEspMetType :=
 Eval hnf in  EspMetType T T numEspMetMixin.
Canonical Structure numTopologyType := 
 Eval hnf in TopologyType T (EspMet_TopMixin numEspMetType).

End nummetrie.

Section canonical.

Canonical Structure numFieldEspMetType (R : numFieldType) := 
  [espMetType R of R for (numEspMetType R)].
Canonical Structure numFieldTopologyType (R : numFieldType) := 
  [topologyType of R for (numTopologyType R)].
Canonical Structure numClosedFieldEspMetType (R : numClosedFieldType) := 
  [espMetType R of R for (numEspMetType R)].
Canonical Structure numClosedFieldTopologyType (R : numClosedFieldType) := 
  [topologyType of R for (numTopologyType R)].
Canonical Structure realEspMetType (R : realDomainType) := 
  [espMetType R of R for (numEspMetType R)].
Canonical Structure realTopologyType (R : realDomainType) := 
  [topologyType of R for (numTopologyType R)].
Canonical Structure realFieldEspMetType (R : realFieldType) := 
  [espMetType R of R for (numEspMetType R)].
Canonical Structure realFieldTopologyType (R : realFieldType) := 
  [topologyType of R for (numTopologyType R)].
Canonical Structure rcfEspMetType (R : rcfType) := 
  [espMetType R of R for (numEspMetType R)].
Canonical Structure rcfTopologyType (R : rcfType) := 
  [topologyType of R for (numTopologyType R)].
Canonical Structure archiDomainEspMetType (R : archiDomainType) := 
  [espMetType R of R for (numEspMetType R)].
Canonical Structure archiDomainTopologyType (R : archiDomainType) := 
  [topologyType of R for (numTopologyType R)].
Canonical Structure archiFieldEspMetType (R : archiFieldType) := 
  [espMetType R of R for (numEspMetType R)].
Canonical Structure archiFieldTopologyType (R : archiFieldType) := 
  [topologyType of R for (numTopologyType R)].

End canonical.

Section Cmetrie.

Variable Rc : rcfType.
Local Notation C := (complex Rc).

Open Scope ring_scope.

Definition dist_C (x y : C) := Re `|x - y|.

Lemma dist_C_ge0 x y : 0 <= dist_C x y.
Proof. by rewrite Re_ge0. Qed.

Lemma dist_C_sym x y : dist_C x y = dist_C y x.
Proof. by rewrite /dist_C distrC. Qed.

Lemma dist_C_refl x y : reflect (x = y) (dist_C x y == 0).
Proof.
rewrite /dist_C real_Re_eq0 // normr_eq0 subr_eq0.
exact: eqP.
Qed.

Lemma dist_C_tri x y z : dist_C x y <= dist_C x z + dist_C z y.
Proof.
rewrite /dist_C -raddfD -{1}(addrNK z x) -addrA.
by rewrite -real_lecE ?realD // ler_norm_add.
Qed.

Canonical Structure EspMet_CMixin :=
  EspMetMixin dist_C_ge0 dist_C_sym dist_C_refl dist_C_tri.
Canonical Structure EspMet_C :=
 Eval hnf in  EspMetType Rc C EspMet_CMixin.
Canonical Structure Esp_Top_C :=
 Eval hnf in TopologyType C (EspMet_TopMixin EspMet_C).

End Cmetrie.

Section numdomain.

Variable R : numDomainType.
Open Scope ring_scope.

Lemma cvgN l (U : nat -> R) :  ((fun n => - (U n)) >->> - l) <-> U >->> l. 
Proof.
split=> /Un_cvgMT H; apply/Un_cvgMT=> r; have [N HN]:= H r; exists N=> n Hn.
  by rewrite insE /= /dist_real -normrN opprD; apply: HN.
by rewrite insE /= /dist_real -opprD normrN; apply: HN.
Qed.

Lemma cvg_addl a l (U : nat -> R) :
   (fun n => a + U n) >->> (a + l) <-> U >->> l .
Proof.
split=> /Un_cvgMT H; apply/Un_cvgMT=> r; have [N HN]:= H r; exists N=> n Hn.
by rewrite insE /= /dist_real -[l](addrK a) [l+a]addrC -addrA -opprD; apply: HN.
by rewrite insE /= /dist_real opprD addrA [a+l]addrC addrK; apply: HN.
Qed.

Lemma cvg_addr a l (U : nat -> R) :
   (fun n => U n + a) >->> (l + a) <-> U >->> l .
Proof.
have ext n : a + U n = U n + a by rewrite addrC.
split=>[/(cvg_ext ext)|/(cvg_addl a) H]; rewrite addrC.
  by move/(cvg_addl).
exact/(cvg_ext (fsym ext)).
Qed.

Lemma cvg_normr0 (f : nat -> R) : (fun k => `|f k|) >->> 0 -> f >->> 0.
Proof.
move/Un_cvgMT => nf0; apply/Un_cvgMT => r; have [N Hn] := nf0 r.
exists N => n nN; have := Hn n nN.
by rewrite !insE /= /dist_real !add0r !normrN normr_id.
Qed.

Lemma ler0_cvg (U V : nat -> R) :
  (forall n, 0 <= U n <= V n) -> (V >->> 0) -> U >->> 0.
Proof.
move=> H /Un_cvgMT cvgV.
apply/Un_cvgMT=> r.
have [k Hk]:= cvgV r. 
exists k=> n Hn. 
rewrite insE /= /dist_real sub0r normrN.
have/andP [Ul0 UlV] := H n.
rewrite (@ler_lt_trans _ `|0 - V n|) // ?Hk // sub0r normrN. 
by rewrite !ger0_norm // (ler_trans Ul0 UlV).
Qed.

Lemma normr_cvg0 (U V : nat -> R) : (forall n, `|V n| <= U n) ->
  U >->> 0 -> V >->> 0.
Proof.
move=> VleU Ucv0.
apply/cvg_normr0/(ler0_cvg _ Ucv0)=> n.  
by rewrite normr_ge0 VleU.
Qed.

Lemma continue_norm : continue (fun (x : R) => `|x|).
Proof.
by apply/continueMT=> x r; exists r=> y; apply/ler_lt_trans/ler_dist_dist.
Qed.


Lemma ler_mulr_cvg0 (U V W : nat -> R) : (forall n, `|U n| <= `|V n|) ->
  (fun n => V n * W n) >->> 0 -> (fun n => U n * W n) >->> 0.
Proof.
move=> leUV /(continue_cvg continue_norm).
rewrite normr0=> VWcv0.
apply/cvg_normr0/(ler0_cvg _ VWcv0)=> n.
by rewrite normr_ge0 /= !normrM ler_wpmul2r.
Qed.

End numdomain.

Section numfield.

Variable R F1 : numFieldType.
Open Scope ring_scope.

Lemma cvg_mulr (a b : R) (U : nat -> R) : U >->> a -> 
(fun k => b * U k) >->> (b * a).  
Proof.
move=> /Un_cvgMT HUa.
case: (altP (b =P 0))=> [->|Hb]; first rewrite mul0r.
  have H1:  (fun k => 0 * U k) =1 (fun _ => 0) by move=> x /=; rewrite mul0r.
  by apply/(cvg_ext H1)/cvg_const.
apply/Un_cvgMT=> r.
have Hpos : 0 < sort r / `|b|.
  by rewrite mulr_gt0 ?lt_pos // invr_gt0 normr_gt0.
have[N HN]:= HUa (PosD Hpos).
exists N=> n Hn.
rewrite insE /= /dist_real -mulrBr normrM mulrC -ltr_pdivl_mulr ?normr_gt0 //.
exact: HN.
Qed.

Lemma cvg_mull (a b : R) (U : nat -> R) : U >->> a -> 
(fun k => U k * b) >->> (a * b).  
Proof.
move=> HUa.
have H1 : (fun k : nat => U k * b) =1 (fun k : nat => b * U k).
  by move=> x /=; rewrite mulrC.
by rewrite mulrC; apply/(cvg_ext H1)/cvg_mulr.
Qed.

Lemma continue2 (F : numFieldType)
 (f : F1 -> F) c (lt0c : 0 < c) x :
continue_pt_met f x <-> forall r, exists r', forall y, (y 'in 'B(x,r')) -> 
(f y) 'in 'B(f x,(PosD (mulr_gt0 (lt_pos r) lt0c))). 
Proof.
split=> Hc r; first exact: Hc.
have pos : 0 < sort r / c by rewrite mulr_gt0 ?lt_pos // invr_gt0.
have [r' Hr'] := Hc (PosD pos).
exists r'=> y Hy; have := Hr' y Hy.
rewrite insE /= /dist_real mulfVK // eqr_le. 
by rewrite real_lerNgt ?lt0c ?real0 // gtr0_real.
Qed.

Lemma continueD (F : numFieldType) (f g : F1 -> F) :
continue f -> continue g -> continue (fun x => f x + g x).
Proof.
move=> /continueMT Hf /continueMT Hg.
have pos2 : ((0 : F) < 2%:R) by rewrite ltr0n.
apply/continueMT=> x; apply/(continue2 _ pos2)=> r.
have [rf Hrf]:= Hf x r.
have [rg Hrg]:= Hg x r.
have pos : 0 < minr rf rg by rewrite minr_gt0 // lt_pos.
exists (PosD pos)=> y; rewrite !insE /= /dist_real => Hy.
rewrite opprD -addrA [X in _ + X]addrC !addrA -[_ - _ + _]addrA [-_+_]addrC.
rewrite (ler_lt_trans (ler_norm_add _ _)) //.
rewrite mulrDr mulr1 ltr_add //.
  by apply/Hrf/(ltr_le_trans Hy)/ler_minl_pos.
by apply/Hrg/(ltr_le_trans Hy)/ler_minr_pos.
Qed.

Lemma continueM (F : numFieldType) (f g : F1 -> F) :
continue f -> continue g -> continue (fun x => f x * g x).
Proof.
move=> /continueMT Hf /continueMT Hg.
apply/continueMT=> a.
have [r0 Hr0] := Hg a (PosD ltr01).
have pos : 0 < (`|g a| +  1) + `|f a|.
  by rewrite ltr_spaddl // ltr_spaddr.
apply/(continue2 _ pos)=> r.
have [r1 Hr1] := Hf a r.
have [r2 Hr2] := Hg a r.
have posmin : 0 < minr r0 (minr r1 r2) by rewrite !minr_gt0 ?lt_pos.
exists (PosD posmin)=> y; rewrite !insE /= /dist_real=> Hy.
have -> : f a * g a - f y * g y = (f a - f y) * g y + (g a - g y) * f a.
  rewrite !mulrDl addrA [X in _ = X]addrC [in X in _ = X]mulrC mulrN.
  by rewrite -[X in -_ + X]addrA addKr mulrC addrC mulNr.
rewrite (ler_lt_trans (ler_norm_add _ _)) // mulrDr !normrM ltr_le_add //.
  apply: ltr_pmul=> //.
    apply/Hr1/(ltr_le_trans Hy).
    by rewrite (ler_trans _ (ler_minl_pos r1 r2)) // ler_minr_pos.
  have/(Hr0 y): y 'in 'B(a,r0) by apply/(ltr_le_trans Hy); rewrite ler_minl_pos.
  rewrite insE /= /dist_real distrC => glt1.
  have := ler_lt_trans (ler_sub_dist (g y) (g a)) glt1.
  by rewrite ltr_subl_addl.
rewrite ler_wpmul2r // ltrW //.
apply/Hr2/(ltr_le_trans Hy).
by rewrite (ler_trans _ (ler_minr_pos r1 r2)) // ler_minr_pos.
Qed.

Lemma continueX (n : nat) : continue (fun (x : F1) => x^+n).
Proof.
elim: n =>[|n IHn]; first exact: continue_const.
set g := (X in continue X).
have ext: (fun x => x * x^+n) =1 g by move=> x; rewrite /g /= exprS.
apply/(continue_ext ext)/(continueM _ IHn).
exact: continue_id.
Qed.

Lemma cvg_add l l' (U V : nat -> F1) : U >->> l -> V >->> l' -> 
  (fun n => U n + V n) >->> (l + l').
Proof.
move=> /Un_cvgMT HU /Un_cvgMT HV; apply/Un_cvgMT=> r.
have [N1 HN1] := HU (PosD (mulr_gt0 (lt_pos r) (inv_INR_pos _ 1%N))).
have [N2 HN2] := HV (PosD (mulr_gt0 (lt_pos r) (inv_INR_pos _ 1%N))).
exists (maxn N1 N2)=> n.
rewrite geq_max=> /andP [HnN1 HnN2]; rewrite insE /= /dist_real.
rewrite opprD addrA -[l + _ - _]addrA [l'+_]addrC addrA -[_+l'-_]addrA.
rewrite (ler_lt_trans (ler_norm_add _ _)) //.
have ->: (sort r) = (sort r) / 2%:R + (sort r) / 2%:R.
  by rewrite -[X in X + X]mul1r -mulrDl mulrC mulrVK // unitfE pnatr_eq0.
by rewrite ltr_add //; [apply: HN1|apply: HN2].
Qed.

End numfield.

Section realdomain.

Variable R : realDomainType.
Open Scope ring_scope. 

Lemma le_cvg_le (a b : R) (U_ : nat -> R) : 
 (forall k, a <= U_ k) -> (U_ >->> b) -> a <= b.
Proof.
move=> Hle /Un_cvgMT Hcvg.
apply: lt_eps_le=> r Hpos.
have [N HN] := (Hcvg (PosD Hpos)).
rewrite (ler_lt_trans (Hle N)) // addrC -[X in X < _](addrNK b).
rewrite ltr_add2r -ltr_opp2 opprB.
have := HN N (leqnn N).
rewrite insE ltr_norml.
by case/andP .
Qed.

Lemma lt_cvg_le (a b : R) (U_ : nat -> R) : 
 (forall k, a < U_ k) -> (U_ >->> b) -> a <= b.
Proof.
move=> Hlt /Un_cvgMT Hcvg.
apply: lt_eps_le=> r Hpos.
have [N HN] := (Hcvg (PosD Hpos)).
rewrite (ltr_trans (Hlt N)) // addrC -[X in X < _](addrNK b).
rewrite ltr_add2r -ltr_opp2 opprB.
have := HN N (leqnn N).
rewrite insE ltr_norml.
by case/andP.
Qed.

Lemma sandwich (U V W : nat -> R) l : (forall n, U n <= V n <= W n) ->
(U >->> l) -> (W >->> l) -> (V >->> l). 
Proof.
move=> Hle /Un_cvgMT HUl /Un_cvgMT HWl.
apply/Un_cvgMT=> r.
have [N1 HN1] := HUl r.
have [N2 HN2] := HWl r.
exists (maxn N1 N2)=> n; rewrite geq_max=> /andP [Hn1 Hn2].
have := HN1 n Hn1.
have := HN2 n Hn2.
have/andP [? ?] := Hle n.
rewrite !insE /= /dist_real !ltr_norml=> /andP [HW _] /andP [_ HU].
by rewrite (ltr_le_trans HW) ?(ler_lt_trans _ HU) // ler_add2l ler_opp2.
Qed.

End realdomain.

Section archi.

Variable R : archiDomainType.
Open Scope ring_scope.

Lemma inv_cvg0 : (fun n => (n.+1%:R^-1 : R)) >->> 0.
Proof.
apply/Un_cvg3=> k; exists k.+1=> n Hn; rewrite insE /= /dist_real sub0r normrN.
by rewrite ger0_norm ?ltrW ?invr_gt0 ?ltr0n // -ltn_ltrV.
Qed.

Lemma cvg_lt1 (x : R) : reflect ((fun k => x^+ k) >->> 0) (`|x| < 1).
Proof.
apply/(iffP idP)=>[xsm|].
  have := normr_ge0 x; rewrite ler_eqVlt.
  case/orP.
    rewrite eq_sym normr_eq0 => /eqP ->.
    apply/(@cvg_shift (archiDomainTopologyType R) 0 [eta GRing.exp 0] 1%N).
    have H: (fun k => 0^+ (k + 1)) =1 (fun _ => (0 : R)).
      by move=> y /=; rewrite addn1 expr0n.
    by apply/(cvg_ext H)/cvg_const.
  move=> xgt0.
  apply/cvg_normr0/Un_cvg3=> n.
  set N := boundD (n.+1%:R * (`|x| ^-1 - 1) ^-1).
  exists N => n' Nn'.
  rewrite insE /= /dist_real add0r normrN normr_id normrX.
  have xXNp : 0 < `|x| ^+ N  by exact: exprn_gt0.
  apply: (ler_lt_trans (_ : _ <= `|x | ^+ N)).
    rewrite -[N]addn0 -(subnKC Nn') !exprD (ler_pmul2l xXNp) expr0.
    by rewrite exprn_ile1 // ltrW.
  rewrite {n' Nn'} -(ltr_pmul2l (ltr0Sn _ n)) mulfV; last first.
    by rewrite pnatr_eq0.
  have xXNnz : `|x| ^+ N != 0. 
    by rewrite eqr_le real_lerNgt ?xXNp ?real0 // gtr0_real.
  rewrite -[X in _ < X](mulVf xXNnz) (ltr_pmul2r (xXNp)).
  rewrite -exprVn (_ : `|x|^-1 = 1 + (`|x|^-1 - 1)); last first.
    by rewrite addrCA addrN addr0.
  have invx1 : 0 < `|x| ^- 1 - 1 by rewrite subr_gt0 invf_gt1.
  have bnd_exp: forall a : R, 0 < a -> 1 + N%:R * a <= (1 + a) ^+ N.
    clear; elim: {n} N; first by move => a ap; rewrite mul0r addr0 expr0.
    move => n Ih a ap.
    rewrite -add1n natrD mulrDl mul1r addrCA exprS mulrDl mul1r addrC.
    rewrite ler_add ?Ih // -{1}[a]mulr1 (ler_pmul2l ap).
    by apply: (ler_trans _ (Ih _ ap)); rewrite cpr_add mulr_ge0 // ?ler0n ?ltrW.
  have nn1 : n.+1%:R < 1 + n.+1%:R :> R by rewrite cpr_add ltr01.
  apply/(ltr_trans nn1)/(ltr_le_trans _ (bnd_exp _ invx1)).
  by rewrite ltr_add2l -ltr_pdivr_mulr ?archiD_boundP ?ltrW ?divr_gt0 ?ltr0Sn.
case/Un_cvgMT/(_ (PosD ltr01))=> N /(_ N.+1 (leqnSn N)).
by rewrite insE /= /dist_real sub0r normrN normrX expr_lt1.
Qed.

Lemma polyM_exp_cvg0 :
forall i (a : R),`|a| < 1 ->
  (fun x => x%:R ^+ i * a ^+ x) >->> 0.
Proof.
move=> i a ha.
apply/(cvg_shift _ _ 1).
set f := (X in X >->> _).
have ext1: f =1 (fun x => x.+1%:R ^+ i * a ^+ x.+1).
  by move=> x; rewrite /f /= !addn1.
apply/(cvg_ext ext1); rewrite /f; clear f ext1; move: ha.
wlog : a / 0 <= a => [wlg Ha|]. 
  have ext: (fun x : nat => `|x.+1%:R ^+ i * a ^+ x.+1|) =1
            (fun n : nat => (n.+1%:R ^+ i * `|a| ^+ n.+1)).
    by  move=> x /=; rewrite normrM !normrX [`|_%:R|]ger0_norm // ler0n.
  by apply/cvg_normr0/(cvg_ext ext)/wlg=> //; rewrite normr_id.
rewrite ler_eqVlt; case/orP=> [/eqP <- _|a0].
  have H1 x : x.+1%:R ^+ i * 0 ^+ x.+1 = (fun _ => (0 : R)) x.
    by rewrite expr0n mulr0.
  by apply/(cvg_ext H1)/cvg_const.
rewrite gtr0_norm // => a1.
have a'p : 0 < (1 - a) / (2%:R * a).
  by rewrite mulr_gt0 ?subr_gt0 // invr_gt0 mulr_gt0 // ltr0n.
have/Un_cvgMT cvgXV: ((fun n => n.+1%:R^+ i / n%:R^+i) : nat -> R) >->> 1.
  apply/(cvg_shift _ _ 1).
  set U := (X in X >->> _).
  pose f := (fun (x : R) => x ^+ i).
  have ext : U =1 f \o (fun n => 1 + n.+1%:R^-1). 
    rewrite /U => n /=.
    rewrite -exprVn -exprMn -[(_+_).+1]addn1 natrD mulrDl.
    by rewrite mulfV ?pnatr_eq0 ?addn1 // mul1r.
  apply/(cvg_ext ext); rewrite -[X in _ >->> X](expr1n _ i).
  apply: continue_cvg; first exact: continueX.
  by rewrite -[X in _ >->> X]addr0; apply/(cvg_addl 1 0)/inv_cvg0.
have [N HN]:= cvgXV (PosD a'p).
have cv0 : (fun n => N.+1%:R ^+ i * a ^+ N.+1 * ((a+1)/2%:R)^+ n) >->> 0.
  rewrite -(mulr0  (N.+1%:R ^+ i * a ^+ N.+1)).
  apply/cvg_mulr/cvg_lt1; rewrite ger0_norm.
    by rewrite ltr_pdivr_mulr ?ltr0n // mul1r ltr_le_add.
  rewrite mulr_ge0 ? invr_ge0 ?ler0n // ltrW //.
  by rewrite (ltr_le_trans a0) // ler_addl ler01.
apply/(cvg_shift _ _ N)/(ler0_cvg _ cv0)=> n. 
rewrite mulr_ge0 ?exprn_ge0 // ?ler0n // ?(ltrW a0) //=.
elim: n =>[|n IHn]; first by rewrite expr0 mulr1.
rewrite [(_/_)^+_]exprSr [_ *(_^+n * _)]mulrA.
set x := (_+_).+1; set y := (n + N).+1.
set xpi := x%:R ^+ i; pose ypi : R := y%:R ^+ i.
have ->: xpi * a ^+ x = ypi * a ^+ y * ((xpi) / (ypi) * a).
  rewrite -mulrA [_^+_ * _]mulrC [_/_]mulrC ![in X in _ = X]mulrA.
  by rewrite mulfV ?expf_neq0 ?pnatr_eq0 // mul1r -mulrA -exprS.
rewrite ler_pmul ?mulr_ge0 ?invr_ge0 ?exprn_ge0 ?ler0n ?[_<= a]ltrW //.
have Hn: (N <= n.+1 + N)%N by rewrite leq_addl. 
have:= HN _ Hn; rewrite insE /= /dist_real /xpi /ypi /x /y.
rewrite -normrN opprB ger0_norm.
  have nz2: 2%:R * a != (0 : R).
    rewrite eqr_le real_lerNgt ?real0 ?realM ?realn ?gtr0_real //.
    by rewrite mulr_gt0 // ltr0n.
  rewrite ltr_subl_addr -[X in _ + X](mulfV nz2) -mulrDl [X in _ + X]mulrDl.
  rewrite mul1r addrA addrNK addrC -(ler_pdivl_mulr _ _ a0) invfM mulrA. 
  exact: ltrW.
rewrite subr_ge0 -exprVn -exprMn exprn_ege1 ?ler_pdivl_mulr ?mul1r ?ler_nat //.
by rewrite ltr0n.
Qed.

End archi.

Section rcf.

Variable R : rcfType.
Open Scope ring_scope.

Lemma cvg_realc (U : nat -> R) :
   U >->> 0 <-> (fun n => (U n)%:C) >->> 0.
Proof.
split=>/Un_cvgMT H; apply/Un_cvgMT=> r; have[N HN]:= H r; exists N=> n Hn.
  rewrite insE /= /dist_C -rmorphB normc_real.
  exact: HN.
have := HN n Hn.
by rewrite insE /= /dist_C /dist_real -rmorphB normc_real.
Qed.

End rcf.

Section Norm.

Section normDef.

Variable R : numDomainType.
Open Scope ring_scope.

Structure can_norm : Type := MNorm {
  anorm : forall p q, 'M[R]_(p,q) -> R;
  pos_norm : forall p q (A : 'M_(p, q)), 0 <= anorm A;
  hom_pos_norm : forall p q (A : 'M_(p, q)) (r : R),
    anorm (r *: A) = `|r| * anorm  A;
  trin_ineq_norm : forall p q (A B : 'M_(p, q)), 
    anorm (A + B) <= anorm A + anorm B;
  prod_ineq_norm : forall p q r (A : 'M_(p, q)) (B : matrix _ q r),
    anorm (A *m B) <= anorm A * anorm B;
  canonical_norm : forall p q (A : 'M_(p, q) ) i j, `|A i j| <= anorm A;
  canonical_norm2 : forall p q (A B: 'M_(p, q)), 
    (forall i j, `|A i j| <= `|B i j|) -> anorm A <= anorm B
}.

End normDef.

Local Notation "|| A : N |" := (anorm N A)
  (at level 0, A, N at level 99, format "|| A  :  N |") : ring_scope.

Section normTheory.

Variable R : numDomainType.
Variable p q: nat.
Variable N : can_norm R.

Open Scope ring_scope.

Lemma cnorm0 : || (0: 'M_(p, q)) : N | = 0.
Proof.
set z := (0 : 'M_(p,q)).
have: || z : N | = || (2%:R *: z) : N | by rewrite scaler0.
by rewrite hom_pos_norm normr_nat mulr_natl -[X in X = _]addr0=> /addrI <-.
Qed.

Lemma cnorm_eq0 (A : 'M[R]_(p,q)) : (|| A : N | == 0) = (A == 0).
Proof.
apply/idP/idP=> [/eqP Heq0|/eqP ->].
  apply/eqP/matrixP=> i j; rewrite mxE.
  by apply/eqP; rewrite -normr_eq0 eqr_le -{1}Heq0 canonical_norm /=.
by rewrite cnorm0.
Qed.
 
Lemma cnorm_lt0  (A: 'M_(p, q)) : (0 < ||A : N |) = (A != 0).
Proof. by rewrite ltr_neqAle pos_norm -cnorm_eq0 eq_sym andbT. Qed.

Lemma cnormN  (A: 'M_(p, q)) : || -A : N| = || A : N|.
Proof. by rewrite -[A]scale1mx -scaleNr !hom_pos_norm normrN. Qed.

End normTheory.
End Norm.

Notation "|| A : N |" := (anorm N A)
  (at level 0, A, N at level 99, format "|| A  :  N |") : ring_scope.

Section Defnorm.

Variable R : numDomainType.

Open Scope ring_scope.

Definition norm8 (m n : nat) (A : 'M[R]_(m,n)) :=
   \maxr_i \sum_j `|A i j|.

Lemma norm8_pos (m n : nat) (A : 'M[R]_(m,n)) : 0 <= norm8 A. 
Proof. by apply: bigmax_pos=> i; apply: sumr_ge0. Qed.

Lemma norm8_hom p q (A : 'M_(p, q)) (r : R) : norm8 (r *: A) = `|r| * norm8 A.
Proof.
rewrite /norm8 (eq_bigr (fun i => `|r| * \sum_j `|A i j|)) ?bigmax_distrl //.
move=> i _; rewrite big_distrr /=.
by apply: eq_bigr=> j _; rewrite mxE normrM.
Qed.

Lemma norm8_tri p q (A B : 'M_(p, q)) : norm8 (A + B) <= norm8 A + norm8 B.
Proof.
have Hpos i: 0 <= \sum_j `|(A + B) i j| by apply: sumr_ge0.
have Hle i: \sum_j `|(A + B) i j| <= \sum_j `|A i j| + \sum_j `|B i j|.
  by rewrite -big_split /= ler_sum // => j _; rewrite mxE ler_norm_add.
apply: (ler_trans (bigmax_ler Hpos Hle)).
by apply: bigmax_ler_sum=> i; apply: sumr_ge0.
Qed.

Lemma norm8M p q r (A : 'M_(p, q)) (B : matrix _ q r) :
  norm8 (A *m B) <= norm8 A * norm8 B.
Proof.
have HABge0 i : 0 <= \sum_j `|(A *m B) i j| by apply: sumr_ge0. 
have HAge0 i : 0 <= \sum_j `|A i j| by apply: sumr_ge0.
have HBge0 i : 0 <= \sum_j `|B i j| by apply: sumr_ge0.
have H: norm8 (A *m B) <= \maxr_i \sum_k `|A i k| * \sum_j `|B k j|.
  apply: (bigmax_ler HABge0)=> i.
  rewrite (eq_bigr (fun k => \sum_ j `|A i k * B k j|))=>[|j _]. 
    by rewrite exchange_big /= ler_sum // => j _; rewrite mxE ler_norm_sum //.
  by rewrite big_distrr; apply: eq_bigr=> k _; rewrite normrM.
apply: (ler_trans H).
apply: ler_trans (bigmax_le_prod2 HAge0 HBge0). 
apply: bigmax_ler=> i; first by apply: sumr_ge0=> j _; rewrite mulr_ge0. 
rewrite -bigmax_distrl // big_distrl /= ler_sum // => j _. 
by rewrite ler_pmul // (bigmax_ger HBge0) // mem_index_enum.
Qed.

Lemma norm8_can1 p q (A : 'M_(p, q) ) i j : `|A i j| <= norm8 A.
Proof.
have Hpos k : 0 <= \sum_j `|A k j| by apply: sumr_ge0.
apply: ler_trans (bigmax_ger Hpos (mem_index_enum i)).
by rewrite (bigD1 j) //= ler_addl sumr_ge0.
Qed.

Lemma norm8_can2 p q (A B: 'M_(p, q)) : 
  (forall i j, `|A i j| <= `|B i j|) -> norm8 A <= norm8 B.
Proof.
move=> Hle.
have Hpos k : 0 <= \sum_j `|A k j| by apply: sumr_ge0.
by apply: (bigmax_ler Hpos)=> i; apply: ler_sum.
Qed.

Canonical Structure can_norm8 := MNorm norm8_pos norm8_hom norm8_tri 
  norm8M norm8_can1 norm8_can2. 
 
End Defnorm.
 

Section matrimetrie.

Variable p q : nat.
Variable R : numDomainType.
Let N:= can_norm8 R.

Open Scope ring_scope.

Definition dist_mat (A B : 'M[R]_(p,q)) := 
  ||(A - B) : N|.

Lemma dist_mat_pos (A B : 'M[R]_(p,q)) : 0 <= dist_mat A B.
Proof. exact: pos_norm. Qed.

Lemma dist_mat_sym (A B : 'M[R]_(p,q)) : dist_mat A B = dist_mat B A.  
Proof. by rewrite /dist_mat -opprB cnormN. Qed.

Lemma dist_mat_refl (A B : 'M[R]_(p,q)) : reflect (A = B) (dist_mat A B == 0).
Proof.
rewrite /dist_mat cnorm_eq0 subr_eq0.
exact: eqP.
Qed.

Lemma dist_mat_tri (A B C : 'M[R]_(p,q)) :
  dist_mat A B <= dist_mat A C + dist_mat C B.
Proof. by rewrite /dist_mat -{1}(addrNK C A) -addrA trin_ineq_norm. Qed.

Canonical Structure EspMet_matMixin := 
   EspMetMixin dist_mat_pos dist_mat_sym dist_mat_refl dist_mat_tri.
Canonical Structure EspMet_mat :=
  EspMetType R ('M[R]_(p,q)) EspMet_matMixin.
Canonical Structure EspTop_mat := 
  Eval hnf in TopologyType ('M[R]_(p,q)) (EspMet_TopMixin EspMet_mat).

End matrimetrie.

Section mxTheory.

Section numdomainmx.

Variable R : numDomainType.
Open Scope ring_scope.

Local Notation N8 := (can_norm8 R).

Lemma cvg_scale0r m n (A :'M[R]_(m,n)) U : A != 0 -> 
   (fun k => (U k) *: A) >->> 0 ->  U >->> 0. 
Proof.
move=> HAnz /Un_cvgMT Hcvg0.
apply/Un_cvgMT=> r. 
have HnAlt0 : 0 < ||A : N8| by rewrite cnorm_lt0.
have Hpos : 0 < (sort r * || A : N8|) by rewrite mulr_gt0 // lt_pos.
have[N HN] := Hcvg0 (PosD Hpos).
exists N=> k Hk.
rewrite insE /= /dist_real sub0r normrN.
rewrite -(ltr_pmul2r HnAlt0) -hom_pos_norm -cnormN -sub0r.
exact: HN.
Qed.

Lemma norm8_col_mx m1 m2 n (A : 'M_(m1,n)) (D : 'M_(m2,n)) :
  || col_mx A D : N8| = maxr ||A : N8| ||D : N8|.
Proof.
case: m1 A=>[A|m1 A].
  by rewrite col_flat_mx [X in maxr X _]/= /norm8 big_ord0 max0r // pos_norm.
have Hspos1 i : 0 <= \sum_j `|A i j| by rewrite sumr_ge0.
have Hnv1: (index_enum (ordinal_finType m1.+1)) != [::].
  by rewrite -size_eq0 /index_enum -enumT size_enum_ord.
have [k1 /andP [_ /eqP Hk1]] := bigmax_eq Hspos1 Hnv1.
case: m2 D=>[D|m2 D].
  rewrite [X in maxr _ X]/= /norm8 big_ord0 /maxr pos_norm //= /norm8.
  have Ho p (k : 'I_p) : (k < p + 0)%N by rewrite addn0.
  have Ho2 p : forall k : 'I_(p + 0), (k < p)%N by rewrite addn0.
  have Heq_sum (k : 'I_(m1.+1 + 0)) (l : 'I_m1.+1) : (k = l :> nat) -> 
     \sum_i `|(col_mx A D) k i| = \sum_i `|A l i|. 
    move=> Heq; apply/eq_bigr=> i _; rewrite mxE; case: splitP=> [j Hj|[][] //].
    by have ->: j = l by apply: ord_inj; rewrite -Hj.
  apply: bigmax_def=>[i||j _]; first by rewrite sumr_ge0.
    exists (Ordinal (Ho _ k1)); rewrite -Hk1 mem_index_enum andbT eq_sym. 
    by apply/eqP/Heq_sum.
  rewrite (Heq_sum _ (Ordinal (Ho2 _ j))) //.
  by apply: (bigmax_ger Hspos1); rewrite mem_index_enum.
have Hspos2 i : 0 <= \sum_j `|D i j| by rewrite sumr_ge0.
have Hnv2: (index_enum (ordinal_finType (m2.+1))) != [::].
  by rewrite -size_eq0 /index_enum -enumT size_enum_ord.
have [k2 /andP [_ /eqP Hk2]] := bigmax_eq Hspos2 Hnv2.
rewrite /= /norm8.
apply: bigmax_def=>[i||j _]; first by rewrite sumr_ge0.
    rewrite/Num.max; case: ifP => _.
    exists (lshift m2.+1 k1); rewrite mem_index_enum -Hk1 andbT.
    by apply/eqP/eq_bigr=> i _; rewrite col_mxEu.    
  exists (rshift m1.+1 k2); rewrite mem_index_enum -Hk2 andbT.
  by apply/eqP/eq_bigr=> i _; rewrite col_mxEd.
case: (ltnP j m1.+1)=> H.
  have -> : j = (lshift _ (Ordinal H)) by apply: ord_inj.
  rewrite (eq_bigr (fun l => `|A (Ordinal H) l|))=>[|i _]; last first.
    by rewrite col_mxEu.
  rewrite real_ler_maxr ?ger0_real ?bigmax_pos //. 
  by rewrite (bigmax_ger Hspos1) // mem_index_enum.
have Ho : (j - m1.+1 < m2.+1)%N.
  by rewrite -(ltn_add2r m1.+1) subnK // [X in (_ < X)%N]addnC.
have ->: j = rshift _ (Ordinal Ho) by apply: ord_inj=> /=; rewrite subnKC.
rewrite (eq_bigr (fun l => `|D (Ordinal Ho) l|))=>[|i _]; last first.
  by rewrite col_mxEd.
rewrite real_ler_maxr ?ger0_real ?bigmax_pos //. 
by rewrite (bigmax_ger Hspos2) ?orbT // mem_index_enum.
Qed.

Lemma norm8_row_mx0l m n1 n2 (A : 'M_(m,n2)) :
  || row_mx (0 : 'M_(m,n1)) A : N8| = ||A : N8|.
Proof.
case: m A=>[A|m A].
  have ->: A = 0 by apply/matrixP=> [][].
  by rewrite row_mx_const !cnorm0.
have Hspos i : 0 <= \sum_j `|A i j| by rewrite sumr_ge0.
have Hnv: (index_enum (ordinal_finType m.+1)) != [::].
  by rewrite -size_eq0 /index_enum -enumT size_enum_ord.
have [k /andP [_ /eqP Hk]] := bigmax_eq Hspos Hnv.
rewrite /= /norm8.
have Heq j : \sum_i `|(row_mx 0 A) j i| = \sum_i `|A j i|.
  rewrite big_split_ord /= big1 ?add0r=> [|i _]; last first.
    by rewrite row_mxEl mxE normr0.    
  rewrite (eq_bigr (fun l => `|A j l|)) // => i _.
  by rewrite row_mxEr.
apply: bigmax_def=>[i||j _]; first by rewrite sumr_ge0.
  by exists k; rewrite mem_index_enum -Hk Heq eqxx.
by rewrite Heq (bigmax_ger Hspos) // mem_index_enum.
Qed.

Lemma norm8_row_mx0r m n1 n2 (A : 'M_(m,n1)) :
  || row_mx A (0 : 'M_(m,n2)) : N8| = ||A : N8|.
Proof.
case: m A=>[A|m A].
  have ->: A = 0 by apply/matrixP=> [][].
  by rewrite row_mx_const !cnorm0.
have Hspos i : 0 <= \sum_j `|A i j| by rewrite sumr_ge0.
have Hnv: (index_enum (ordinal_finType m.+1)) != [::].
  by rewrite -size_eq0 /index_enum -enumT size_enum_ord.
have [k /andP [_ /eqP Hk]] := bigmax_eq Hspos Hnv.
rewrite /= /norm8.
have Heq j : \sum_i `|(row_mx A 0) j i| = \sum_i `|A j i|.
  rewrite big_split_ord /= [X in _ + X]big1 ?addr0=> [|i _]; last first.
    by rewrite row_mxEr mxE normr0.    
  rewrite (eq_bigr (fun l => `|A j l|)) // => i _.
  by rewrite row_mxEl.
apply: bigmax_def=>[i||j _]; first by rewrite sumr_ge0.
  by exists k; rewrite mem_index_enum -Hk Heq eqxx.
by rewrite Heq (bigmax_ger Hspos) // mem_index_enum.
Qed.

Lemma norm8_blockmxul m1 m2 n1 n2 (A : 'M_(m1,n1)) :
  || block_mx A 0 0 (0 : 'M_(m2,n2)) : N8| = ||A : N8|.
Proof. 
rewrite /block_mx norm8_col_mx !norm8_row_mx0r. 
by rewrite /maxr cnorm0 pos_norm.
Qed.

Lemma norm8_blockmxdr m1 m2 n1 n2 (A : 'M_(m2,n2)) :
  || block_mx (0 : 'M_(m1,n1)) 0 0 A : N8| = ||A : N8|.
Proof.
rewrite /block_mx norm8_col_mx !norm8_row_mx0l.
by rewrite cnorm0 max0r // pos_norm.
Qed.

Lemma norm8_blockmxuldr m1 m2 n1 n2 (A : 'M_(m1,n1)) (D : 'M_(m2,n2)) :
  || block_mx A 0 0 D : N8| = maxr ||A : N8| ||D : N8|.
Proof. by rewrite /block_mx norm8_col_mx norm8_row_mx0l norm8_row_mx0r. Qed.

Lemma blockmx_cvgul m1 m2 n1 n2 (U : nat -> 'M[R]_(m1,n1)) (A : 'M_(m1,n1))
 B C (D : 'M_(m2,n2)) :
 (U >->> A) -> (fun k => block_mx (U k) B C D) >->> (block_mx A B C D).
Proof.
move/Un_cvgMT=> HUA; apply/Un_cvgMT=> r.
have [N HN]:= HUA r.
exists N=> n Hn.
rewrite insE /= /dist_mat opp_block_mx add_block_mx !subrr.
by rewrite norm8_blockmxul HN.
Qed.

Lemma blockmx_cvgdr m1 m2 n1 n2 (U : nat -> 'M[R]_(m2,n2)) (A : 'M_(m1,n1))
 B C (D : 'M_(m2,n2)) :
 (U >->> D) -> (fun k => block_mx A B C (U k)) >->> (block_mx A B C D).
Proof.
move/Un_cvgMT=> HUA; apply/Un_cvgMT=> r.
have [N HN]:= HUA r.
exists N=> n Hn.
rewrite insE /= /dist_mat opp_block_mx add_block_mx !subrr.
by rewrite norm8_blockmxdr HN.
Qed.

Lemma blockmx_cvguldr m1 m2 n1 n2 (U : nat -> 'M[R]_(m1,n1))
 (V : nat -> 'M[R]_(m2,n2)) (A : 'M_(m1,n1)) B C (D : 'M_(m2,n2)) :
 (U >->> A) -> (V >->> D) -> 
 (fun k => block_mx (U k) B C (V k)) >->> (block_mx A B C D).
Proof.
move=>/Un_cvgMT HUA /Un_cvgMT HVD.
apply/Un_cvgMT=> r.
have [N1 HN1]:= HUA r.
have [N2 HN2]:= HVD r.
exists (maxn N1 N2)=> n Hn.
rewrite insE /= /dist_mat.
rewrite opp_block_mx add_block_mx !subrr norm8_blockmxuldr.
 rewrite real_ltr_maxl ?ger0_real ?pos_norm ?ltrW //.
by rewrite HN1 ?HN2 // (leq_trans _ Hn) // ?leq_maxl // leq_maxr.
Qed.

Lemma diag_block_mx_cvg s (UF : nat -> forall n, nat -> 'M[R]_n.+1)
  (F : forall n, nat -> 'M[R]_n.+1) :
 (forall i, (i < size s)%N -> 
  (fun k => UF k (nth 0%N s i) i) >->> (F (nth 0%N s i) i))
 -> (fun k => diag_block_mx s (UF k)) >->> (diag_block_mx s F).
Proof.
case: s => [_| a l]; first exact: cvg_const.
elim: l a F UF=> [a F UF H|a l /= IHl b F UF H]; first exact: (H 0%N).
apply: (blockmx_cvguldr (U:= fun k => UF k b 0%N)); first exact: (H 0%N).
by apply: IHl => i Hi; exact: (H i.+1).
Qed.
 
Lemma conform_mx_cvg m n p q B (U : nat -> 'M[R]_(m,n)) 
  (V : nat -> 'M[R]_(p,q)) A :
 m = p -> n = q -> U >->> A -> 
 (fun k => conform_mx (V k) (U k)) >->> (conform_mx B A). 
Proof.
move=> eqmp eqnq; move: U A.
rewrite eqmp eqnq=> U A HU. 
rewrite conform_mx_id.
have ext k : conform_mx (V k) (U k) = (U k) by rewrite conform_mx_id.
by apply/(cvg_ext ext).
Qed.

End numdomainmx.

Section numfieldmx.

Variable R : numFieldType.
Open Scope ring_scope.

Local Notation N8 := (can_norm8 R).

Lemma mx_cvgP m n (U_ : nat -> 'M_(m,n)) (A :'M[R]_(m,n)) :
  U_ >->> A <-> (forall i j, (fun n => (U_ n) i j) >->> (A i j)).
Proof.
split=> [/Un_cvgMT HUcvA i j|Hcv_coeff]; apply/Un_cvgMT=> r.
  case: (HUcvA r)=> N HN.
  exists N => k Hk.
  have Hds: d (A i j) ((U_ k) i j) <= \sum_l (d (A i l) ((U_ k) i l)).
    by rewrite (bigD1 j) // ler_addl sumr_ge0 // => l _; rewrite dist_pos.
  have : \sum_l (d (A i l) ((U_ k) i l)) <= ||(A - U_ k) : N8|. 
    have H p :  0 <= \sum_l (d (A p l) ((U_ k) p l)).
      by rewrite sumr_ge0 // => l _; rewrite dist_pos.
    have Hsm:= bigmax_ger H (mem_index_enum i).
    rewrite (ler_trans Hsm) // (bigmax_ler H) // => l.
    by rewrite ler_sum // => p _; rewrite !mxE.
   move/(ler_trans Hds)=> H.
   exact: (ler_lt_trans _ (HN k Hk)).
have {Hcv_coeff} Hcvcf i j : Un_cvg_met (fun n0 : nat => (U_ n0) i j) (A i j).
 by apply/Un_cvgMT.
rewrite /open_ball.
case: n U_ A r Hcvcf=> [U_ A r Hcvcf|n U_ A r Hcvcf].
  exists 0%N=> n _; rewrite insE.
  have/dist_refl/eqP -> : A = U_ n by apply/matrixP=> i [][].
  by rewrite lt_pos.
case: m U_ A r Hcvcf=> [U_ A r Hcvcf|m U_ A r Hcvcf].
  exists 0%N=> m _; rewrite insE.
  have/dist_refl/eqP -> : A = U_ m by apply/matrixP=> [][][].
  by rewrite lt_pos.
have Hpos: 0 < (sort r) / n.+1%:R by rewrite mulr_gt0 ?lt_pos // inv_INR_pos.
pose N := \max_i (\max_j pi1 (cid (Hcvcf i j (PosD Hpos)))).
exists N=> k Hk.
have Hs i :  0 <= \sum_j `|(A - (U_ k)) i j|.
  by rewrite sumr_ge0.
have Hnnul : index_enum (ordinal_finType m.+1) != [::].
 by rewrite -size_eq0 /index_enum -enumT size_enum_ord.
have [i /andP [_ /eqP Hi]]:= bigmax_eq Hs Hnnul.
rewrite insE /= /dist_mat /= /norm8 -Hi.
have Hu: (n.+1%:R : R) \is a GRing.unit by rewrite unitfE pnatr_eq0. 
rewrite -[sort r](mulrVK Hu) mulr_natr -[X in _*+ X]card_ord -sumr_const.
apply: (ltr_sum _ (mem_index_enum (ord0 : 'I_n.+1)))=> // j _; rewrite !mxE.
apply/(pi2 (cid (Hcvcf i j (PosD Hpos))))/(leq_trans _ Hk).
apply: (leq_trans (leq_bigmax (F := (fun j => pi1 (cid (Hcvcf i j _)))) j)).
exact: (leq_bigmax (F := (fun i => \max_j pi1 (cid (Hcvcf i j _)))) i).
Qed.

Lemma continu_mulr : forall m n p (A : 'M[R]_(m,n)),
   continue (fun (B : 'M[R]_(n,p)) => A *m B).
Proof.
move=> m n p A; apply/continueMT=> B r.
have pos: 0 < (sort r) / ( ||A : N8| + 1).
  by rewrite divr_gt0 // ltr_spaddr // pos_norm.
pose r2 := PosD pos.
exists r2=> C HC.
rewrite insE /= /dist_mat -mulmxBr.
apply: (ler_lt_trans (prod_ineq_norm _ A (B - C))).
apply: (ler_lt_trans (y:= ||A : N8| * (sort r2))).
  by rewrite ler_wpmul2l ?pos_norm // ltrW.
rewrite /r2 /=.
rewrite mulrC -mulrA gtr_pmulr ?lt_pos //.
have Hgt0 : 0< (norm8 A + 1).
  by rewrite -[0]addr0 ler_lt_add ?ltr01 // pos_norm.
rewrite -(ltr_pmul2l Hgt0) mulrA mulrV ?mul1r ?mulr1 ?unitf_gt0 //.
by rewrite -[X in X < _]addr0 ltr_add2l ltr01.
Qed.

Lemma continu_mull : forall m n p (A : 'M[R]_(n,p)),
    continue (fun (B : 'M[R]_(m,n)) => B *m A ).
Proof.
move=> m n p A; apply/continueMT=> B r.
have pos: 0 < (sort r) / ( ||A : N8| + 1).
  by rewrite divr_gt0 // ltr_spaddr // pos_norm.
pose r2 := PosD pos.
exists r2=> C HC.
rewrite insE /= /dist_mat -mulmxBl.
apply: (ler_lt_trans (prod_ineq_norm _ (B - C) A)).
apply: (ler_lt_trans (y:= (sort r2) * ||A : N8|)).
  by rewrite ler_wpmul2r ?pos_norm // ltrW.
rewrite /r2 /=.
rewrite -mulrA gtr_pmulr ?lt_pos //.
have Hgt0 : 0< (norm8 A + 1).
  by rewrite -[0]addr0 ler_lt_add ?ltr01 // pos_norm.
rewrite -(ltr_pmul2l Hgt0) mulrA mulrV ?mul1r ?mulr1 ?unitf_gt0 //.
by rewrite -[X in X < _]addr0 ltr_add2l ltr01.
Qed.

End numfieldmx.

Section realfieldmx.

Variable R : realFieldType.
Local Notation N8 := (can_norm8 R).
Open Scope ring_scope.
 
Lemma lemx_cvg_lemx m n (A B :'M[R]_(m,n)) (U_ : nat -> 'M_(m,n)) :
 (forall k, A <=m: U_ k) -> (U_ >->> B) -> A <=m: B.
Proof.
move=> Hlemx /mx_cvgP Hcvg i j.
have/le_cvg_le Hle k: A i j <= (U_ k) i j by apply: Hlemx.
exact: Hle.
Qed.

Lemma ltmx_cvg_lemx m n (A B :'M[R]_(m,n)) (U_ : nat -> 'M_(m,n)) :
 (forall k, A <m: U_ k) -> (U_ >->> B) -> A <=m: B.
Proof.
move=> Hltmx /mx_cvgP Hcvg i j.
have/lt_cvg_le Hlt k: A i j < (U_ k) i j by apply: Hltmx.
exact: Hlt.
Qed.

End realfieldmx.

Section archidomainmx.

Variable R : archiDomainType.
Open Scope ring_scope.

Lemma cvg_jordan_lt1 (lam : R) n : `|lam| < 1 -> 
  (fun k => (Jordan_block lam n.+1)^+ k) >->> 0.
Proof.
move=> Hlt1.
apply/(cvg_ext ((Jordan_expn lam n)))/mx_cvgP=> i j; rewrite mxE.
set f := (X in X >->> _).
case: (altP (lam =P 0))=> lam0.
  apply/(cvg_shift _ _ (j - i).+1).
  have ext: (fun i0 : nat => f (i0 + (j - i).+1)%N) =1 (fun _ => 0).
    move=> x; rewrite /f /= !mxE lam0 addnS subSn ?leq_addl // addnK.
    by rewrite expr0n mulr0 mul0rn.
  by apply/(cvg_ext ext)/cvg_const.
case: (leqP i j)=> iltej; last first.
  have ext: f =1 (fun _ => 0) by move=> x; rewrite /f /= !mxE leqNgt iltej.
  by apply/(cvg_ext ext)/cvg_const. 
have cvg0 k : (fun x => ((x + k)%:R^+ k / (k`!%:R)) * lam ^+ x) >->> 0.
  set g := (X in X >->> _).
  have ext x : g x = (x + k)%:R^+ k / (k`!%:R) * lam ^+ (x + k) / lam ^+ k.
    by rewrite -mulrA -exprB ?leq_addl ?unitfE // addnK.
  rewrite -(mul0r (lam^+k)^-1).
  apply/(cvg_ext ext)/cvg_mull; clear ext.
  have ext m: m%:R ^+ k / k`!%:R * lam ^+ m =(m%:R ^+ k * lam ^+ m) / k`!%:R.
    by rewrite -mulrA [_^-1*_^+m]mulrC mulrA. 
  rewrite -(mul0r (k`!%:R)^-1).
  pose U := (fun x => x%:R^+ k / k`!%:R * lam ^+ x). 
  by apply/(iffLR (cvg_shift _ U k))/(cvg_ext ext)/cvg_mull/polyM_exp_cvg0.
apply/(cvg_shift _ _ (j - i)%N).
set g := (X in X >->> _).
have ext: g =1 (fun x => (('C(x + (j - i), j - i))%:R * lam ^+ x)).
  by move=> x; rewrite /g /f /= !mxE iltej addnK.
apply/(cvg_ext ext)/(ler_mulr_cvg0 _ (cvg0 (j - i)%N))=> k.  
rewrite !ger0_norm ?ler0n // ?mulr_ge0 ?invr_ge0 ?exprn_ge0 ?ler0n //.
rewrite ler_pdivl_mulr ?ltr0n ?fact_gt0 // -natrM -natrX ler_nat bin_ffact.
elim: (j - i)%N=> // m IHm.
rewrite expnS addnS ffactSS leq_mul2l orFb (leq_trans IHm) //.
by case: m {IHm}=> // m; rewrite leq_exp2r.
Qed.

End archidomainmx.


Section rcfmx.

Variable R : rcfType.

Local Notation "A ^%:C" := (map_mx (real_complex_def (Phant R)) A) (at level 2).
Local Notation C := (complex R).
Local Notation N8 := (can_norm8 R) .
Local Notation NC := (can_norm8 [numDomainType of C]). 

Open Scope ring_scope.

Lemma normc_mx m n (A : 'M_(m,n)) : ||A ^%:C : NC| = ||A : N8|%:C.
Proof.
rewrite /= /norm8 map_bigmax //.
 +apply: eq_bigr=> i _; rewrite rmorph_sum; apply: eq_bigr=> j _.
 by rewrite !mxE normc_real.
 +by move=> i; rewrite sumr_ge0.
by move=> x y; rewrite lecR.
Qed.
 
Lemma cvg_realc_mx m n (U : nat -> 'M_(m,n)) :
 U >->> 0 <-> (fun (n : nat) => (U n)^%:C) >->> 0.
Proof.
split=>/Un_cvgMT H; apply/Un_cvgMT=> r.
  have Rer: 0 < Re (sort r) by rewrite Re_gt0 // gtr0_real.
  have[N HN]:= H (PosD Rer); exists N=> k Hk.
  have := HN k Hk; rewrite !insE /= /dist_mat !sub0r !cnormN normc_mx.
  by rewrite -{2}[sort r]ReK ?gtr0_real // ltcR.
have Rer: 0 < (sort r)%:C by rewrite ltcR lt_pos.
have[N HN]:= H (PosD Rer); exists N=> k Hk.
have := HN k Hk.
by rewrite !insE /= /dist_mat !sub0r !cnormN normc_mx ltcR.
Qed.

End rcfmx.

End mxTheory.
